from kevin_toolbox.data_flow.file.markdown.table import Table_Format


def get_format(content_s):
    res = None
    if isinstance(content_s, dict):
        if "orientation" in content_s and isinstance(content_s["orientation"], str):
            res = Table_Format.MATRIX
        elif len(content_s) > 0:
            v = list(content_s.values())[0]  # 是 get_format 而不是 check_format，所以只取第一个值进行判断就够了
            if isinstance(v, dict):
                res = Table_Format.COMPLETE_DICT
            elif isinstance(v, (list, tuple)):
                res = Table_Format.SIMPLE_DICT
    return res



import re


def find_links(text, b_compact_format=True, type_ls=None):
    """
        查找文本中的链接

        参数:
            text:               <str> 文本
            b_compact_format:   <bool> 是否只返回 link 部分
                                    默认为 True，此时返回 link_ls，其中每个元素是一个链接
                                    当设置为 False，此时返回 (link_ls, part_slices_ls, link_idx_ls)，
                                        其中 part_slices_ls 是链接和链接前后文本在 text 中对应的 slice，
                                        而 link_idx_ls 指出了 part_slices_ls 中第几个元素对应的是链接，
                                        link_idx_ls 与 link_ls 依次对应。
            type_ls:              <list of str> 找出哪种类型的链接
                                    默认为 None，此时表示找出所有类型的链接。
                                    支持以下取值：
                                        "url", "image"
    """

    matches = re.finditer(r'\[(.*?)\]\((.*?)(?:\s*["\'](.*?)["\'])?\)', text, re.DOTALL)

    link_ls = []
    part_slices_ls = []
    link_idx_ls = []
    start = 0
    for match in matches:
        link_start, link_end = match.start(), match.end()
        #
        if text[link_start - 1] == "!":
            type_ = "image"
            link_start -= 1
        else:
            type_ = "url"
        #
        if type_ls is not None and type_ not in type_ls:
            continue
        #
        part_slices_ls.append([start, link_start])
        # 图片本身
        link_s = dict(
            type_=type_,
            name=match.group(1),
            target=match.group(2),
            title=match.group(3) if match.group(3) else None
        )
        link_idx_ls.append(len(part_slices_ls))
        link_ls.append(link_s)
        part_slices_ls.append([link_start, link_end])
        # 更新起始位置
        start = match.end()

    last = text[start:]
    if last:
        part_slices_ls.append([start, len(text)])

    if b_compact_format:
        return link_ls
    else:
        return link_ls, part_slices_ls, link_idx_ls


if __name__ == "__main__":
    markdown_text = """
    Here is an image:
    ![This is a picture of a cat](http://example.com/cat.jpg "A cute cat")
    And another one:
    ![This is a picture of a dog](http://example.com/dog.jpg 'A cute dog')
    And one without alt text:
    [](http://example.com/placeholder.jpg)
    And one without title:
    ![<image_name>](<image_path>)
    """
    from kevin_toolbox.data_flow.file import markdown

    print(markdown.generate_list(find_links(text=markdown_text, b_compact_format=True)))

    link_ls_, part_slices_ls_, link_idx_ls_ = find_links(text=markdown_text, b_compact_format=False, type_ls=["url"])

    print(link_ls_)
    for part_slices in part_slices_ls_:
        print(part_slices)
        print(markdown_text[part_slices[0]:part_slices[1]])



import time
from .strategy import Strategy_Base
from .cache import Cache_Base
from .variable import CACHE_STRATEGY_REGISTRY, CACHE_BUILDER_REGISTRY


class Cache_Manager:
    """
        缓存管理器

        提供以下接口：
            - 添加条目 add(key, value, b_allow_overwrite)
            - 获取条目 get(key, b_add_if_not_found, default_factory, default)
            - 删除并返回条目 pop(key)
            - 判断是否有该条目 has()
            - 清空所有内容 clear()

        并支持以下用法：
            通过 len(.) 获取缓存大小，通过 in 操作符判断是否有某个条目


        相关变量：
            self.cache：     缓存
            self.metadata_s：缓存的属性数据
                                包含以下字段，各字段将自动更新
                                例如：{
                                        <key>:  {
                                            "last_time": xxx,      # 最近读取时间
                                            "initial_time": xxx,   # 最初读取时间
                                            "survival_time": xxx,  # survival_time:=last_time-initial_time
                                            "counts": xxx,         # 读取次数
                                        },
                                       ...
                                    }
            self.strategy：  缓存更新策略
    """

    def __init__(self, **kwargs):
        """
            参数：
                upper_bound:            <int> 当缓存容量超过该值时，触发重整（抛弃部分条目以将容量降低到 refactor_size 指定的大小）
                refactor_size:          <int/float> 内存重整后的容量大小
                                            当设置为 int 时，表示具体容量的大小；
                                            当设置为 float 时表示占 upper_bound 的比例
                                            默认为 0.5
                strategy:               <str/dict/Strategy_Base> 管理策略
                cache:                  <str/dict/Cache_Base> 缓存种类
        """
        # 默认参数
        paras = {
            "upper_bound": None,
            "refactor_size": 0.5,
            "strategy": ":by_last_time:LRU",
            "cache": ":in_memory:Memo",
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["upper_bound"], int) and paras["upper_bound"] > 0, \
            "upper_bound must be a positive integer"
        if isinstance(paras["refactor_size"], float):
            paras["refactor_size"] = int(paras["refactor_size"] * paras["upper_bound"])
        assert 0 <= paras["refactor_size"] <= paras["upper_bound"], \
            "refactor_size must be less than upper_bound and bigger than zero"
        # strategy
        assert isinstance(paras["strategy"], (str, dict, Strategy_Base)), \
            "strategy must be a string, dict of paras or a Strategy_Base object"
        if isinstance(paras["strategy"], str):
            strategy = CACHE_STRATEGY_REGISTRY.get(name=paras["strategy"])()
        elif isinstance(paras["strategy"], dict):
            strategy = CACHE_STRATEGY_REGISTRY.get(name=paras["strategy"]["name"])(
                **paras["strategy"].get("paras", dict()))
        else:
            strategy = paras["strategy"]
        # cache
        assert isinstance(paras["cache"], (str, dict, Cache_Base)), \
            "cache must be a string, dict of paras or a Cache_Base object"
        if isinstance(paras["cache"], str):
            cache = CACHE_BUILDER_REGISTRY.get(name=paras["cache"])()
        elif isinstance(paras["cache"], dict):
            cache = CACHE_BUILDER_REGISTRY.get(name=paras["cache"]["name"])(**paras["cache"].get("paras", dict()))
        else:
            cache = paras["cache"]

        self.paras = paras
        self.strategy = strategy  # type:Strategy_Base
        self.cache = cache  # type:Cache_Base
        self.metadata_s = dict()  # 保存条目的相关信息

    def add(self, key, value, b_allow_overwrite=True):
        """
            以 key 为查询键，将 value 添加到缓存中

            参数:
                key:
                value:
                b_allow_overwrite:      <boolean> 是否允许覆写缓存中已有的条目
                                            默认为 True
        """
        # 判断是否需要进行覆写
        if self.cache.has(key=key):
            if id(value) != id(self.cache.read(key=key)):
                if b_allow_overwrite:
                    self.__remove_of_cache(key=key)
                else:
                    raise KeyError(f'key: {key} already exists, modification of existing entries is prohibited')
            else:
                return
        # 记录到缓存
        self.__write_of_cache(key=key, value=value)

    def __write_of_cache(self, key, value):
        """
            向缓存中新增不存在的条目
        """
        temp = time.time()
        metadata = {
            "initial_time": temp,  # 最初读取时间（加入缓存的时间）
            "last_time": temp,  # 最近读取时间
            "survival_time": 0.0,  # 最后一次读取与加入缓存时间之差，last_time - initial_time
            "counts": 0,  # 读取次数
        }
        #
        self.cache.write(key=key, value=value)
        self.metadata_s[key] = metadata
        # 通知策略管理器
        self.strategy.notified_by_write_of_cache(key=key, value=value, metadata=metadata)

        # 判断是否重构
        if len(self.cache) <= self.paras["upper_bound"]:
            return
        # 从策略管理器获取建议
        #   根据建议删除条目
        suggest_keys = self.strategy.suggest(refactor_size=self.paras["refactor_size"])
        assert len(suggest_keys) == len(self.cache) - self.paras["refactor_size"], \
            f'expect {len(self.cache) - self.paras["refactor_size"]} deletion suggestions, but got {len(suggest_keys)}'
        for key in suggest_keys:
            self.__remove_of_cache(key=key)

    def __remove_of_cache(self, key):
        """
            从缓存中删除存在的条目
        """
        self.cache.remove(key=key)
        metadata = self.metadata_s.pop(key)
        # 通知策略管理器
        self.strategy.notified_by_remove_of_cache(key=key, metadata=metadata)

    def get(self, key, b_add_if_not_found=False, default_factory=None, **kwargs):
        """
            获取 key 对应的值

            参数:
                key:                    <hashable>
                default:                默认值
                                            当 key 在缓存中不存在时返回
                default_factory:        <callable> 用于产生默认值的函数
                                            当 key 在缓存中不存在时返回该函数的结果
                                            使用该参数相对于 default 能延迟默认值的产生（如果不需要用到默认值就不生成），提高效率
                        注意！！上面两个参数同时指定时，将以前者为准。
                b_add_if_not_found:     <boolean> 当 key 在缓存中不存在时，是否添加到缓存中
                                            默认为 False
        """
        if self.cache.has(key=key):
            value = self.__read_of_cache(key=key)
        elif "default" in kwargs or callable(default_factory):
            value = kwargs["default"] if "default" in kwargs else default_factory()
            if b_add_if_not_found:
                self.__write_of_cache(key=key, value=value)
        else:
            raise KeyError(key)
        return value

    def pop(self, key):
        """
            从缓存中删除 key 对应的条目，并返回该条目的值
        """
        if self.cache.has(key=key):
            value = self.__read_of_cache(key=key)
            self.__remove_of_cache(key=key)
        else:
            raise KeyError(key)
        return value

    def __read_of_cache(self, key):
        """
            读取缓存中 已经存在的 条目
        """
        value = self.cache.read(key=key)
        metadata = self.metadata_s[key]
        # 更新 metadata
        metadata["last_time"] = time.time()
        metadata["survival_time"] = metadata["last_time"] - metadata["initial_time"]
        metadata["counts"] += 1
        # 通知策略管理器
        self.strategy.notified_by_read_of_cache(key=key, value=value, metadata=metadata)

        return value

    def has(self, key):
        """
            判断 key 是否在缓存中
                注意！！不会更新 metadata
        """
        return self.cache.has(key=key)

    def clear(self):
        self.cache.clear()
        self.metadata_s.clear()
        self.strategy.notified_by_clear_of_cache()

    def __len__(self):
        return len(self.cache)

    def __contains__(self, key):
        return self.has(key)



class Saved_Node_Name_Builder:
    """
        生成保存节点内容时的文件夹/文件名称
    """

    def __init__(self, format_):
        try:
            temp = format_.format(**{k: k + "_" * 3 for k in {"raw_name", "id", "hash_name", "count"}})
            assert len(temp) > len(format_)
        except:
            raise ValueError(f'invalid saved_node_name_format {format_}')

        self.format_ = format_
        self.count = 0

    def __call__(self, name, value):
        from kevin_toolbox.nested_dict_list import get_hash

        res = self.format_.format(
            **{"raw_name": name, "id": id(value), "hash_name": get_hash(name, length=12), "count": self.count})
        self.count += 1
        return res


if __name__ == '__main__':
    bd = Saved_Node_Name_Builder(format_="{raw_name}_{count}_{hash_name}_{id}")
    print(bd(":a@0", 1))
    print(bd(":b:c", []))

    # bd = Saved_Node_Name_Builder(format_="")
    # bd = Saved_Node_Name_Builder(format_="{raw_name2}_{count}")



from kevin_toolbox.patches.for_matplotlib.color import Color_Format, convert_format
from kevin_toolbox.patches.for_numpy import random

PREDEFINED = ['blue', 'red', 'green', 'orange', 'purple', 'yellow', "brown", "pink", "gray", "olive", "cyan"]
PREDEFINED = [convert_format(var=i, output_format=Color_Format.HEX_STR) for i in PREDEFINED]

population = tuple('0123456789ABCDEF')


def generate_color_list(nums, seed=None, rng=None, exclude_ls=None, output_format=Color_Format.HEX_STR):
    """
        生成颜色列表

        参数:
            nums:           <int> 生成颜色的数量
            seed,rng:       随机种子或随机生成器，二选一
            exclude:        <list of str> 需要排除的颜色
            output_format:  <Color_Format/str> 输出格式
                            支持 HEX_STR、RGBA_ARRAY 两种格式
        返回：
            不包含 alpha 透明度值的颜色列表
    """
    global PREDEFINED, population
    assert output_format in [Color_Format.HEX_STR, Color_Format.RGBA_ARRAY]
    output_format = Color_Format(output_format)
    if exclude_ls is None:
        exclude_ls = []
    assert isinstance(exclude_ls, (list, tuple))
    exclude_ls = set(convert_format(var=i, output_format=Color_Format.HEX_STR) for i in exclude_ls)
    rng = random.get_rng(seed=seed, rng=rng)

    colors = [i for i in PREDEFINED if i not in exclude_ls][:nums]  # 优先输出预定义的颜色

    # 随机生成剩余数量的颜色
    while len(colors) < nums:
        c = "#" + ''.join(
            rng.choice(population, size=6, replace=True))
        if c not in colors and c not in exclude_ls:
            colors.append(c)
    colors = [convert_format(c, output_format=output_format) for c in colors]

    return colors


if __name__ == '__main__':
    color_list = generate_color_list(1, exclude_ls=['blue'])
    print(color_list)

    color_list = generate_color_list(nums=1, seed=114, exclude_ls=['#0000FF'])
    print(color_list)



import os
import copy
import matplotlib.pyplot as plt
from kevin_toolbox.patches.for_os.path import replace_illegal_chars
from kevin_toolbox.patches.for_matplotlib.color import generate_color_list


def plot_lines(data_s, title, x_name, output_dir=None, **kwargs):
    data_s = copy.copy(data_s)
    line_nums = len(data_s) - 1
    paras = {
        "dpi": 200,
        "color_ls": generate_color_list(nums=line_nums),
        "marker_ls": None,
        "linestyle_ls": '-',
    }
    paras.update(kwargs)
    for k, v in paras.items():
        if k.endswith("_ls") and not isinstance(v, (list, tuple)):
            paras[k] = [v] * line_nums
    assert line_nums == len(paras["color_ls"]) == len(paras["marker_ls"]) == len(paras["linestyle_ls"])

    plt.clf()
    #
    x_all_ls = data_s.pop(x_name)
    data_s, temp = dict(), data_s
    for k, v_ls in temp.items():
        y_ls, x_ls = [], []
        for x, v in zip(x_all_ls, v_ls):
            if x is None or v is None:
                continue
            x_ls.append(x)
            y_ls.append(v)
        if len(x_ls) == 0:
            continue
        data_s[k] = (x_ls, y_ls)
    #
    for i, (k, (x_ls, y_ls)) in enumerate(data_s.items()):
        plt.plot(x_ls, y_ls, label=f'{k}', color=paras["color_ls"][i], marker=paras["marker_ls"][i],
                 linestyle=paras["linestyle_ls"][i])
    plt.xlabel(f'{x_name}')
    plt.ylabel('value')
    plt.title(f'{title}')
    # 显示图例
    plt.legend()

    if output_dir is None:
        plt.show()
        return None
    else:
        # 对非法字符进行替换
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, f'{replace_illegal_chars(title)}.png')
        plt.savefig(output_path, dpi=paras["dpi"])
        return output_path


if __name__ == '__main__':
    plot_lines(data_s={
        'a': [1, 2, 3, 4, 5],
        'b': [5, 4, 3, 2, 1],
        'c': [1, 2, 3, 4, 5]},
        title='test', x_name='a', output_dir=os.path.join(os.path.dirname(__file__), "temp"))



from kevin_toolbox.patches.for_matplotlib.color import Color_Format


def get_format(var):
    if isinstance(var, str):
        if var.startswith("#"):
            res = Color_Format.HEX_STR
        else:
            res = Color_Format.NATURAL_NAME
    else:
        res = Color_Format.RGBA_ARRAY
    return res



import pytest
import time
import random
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.computer_science.algorithm.cache_manager import Cache_Manager


def test_cache_manager_with_lru_strategy():
    print("test Cache_Manager with LRU_Strategy")

    strategy = ":by_last_time:LRU"
    for cache in [":in_memory:Memo", ]:
        cache_manager = Cache_Manager(upper_bound=3, refactor_size=2, strategy=strategy, cache=cache)

        # 添加数据 a、b、c
        #   优先级依次为 c、b、a
        cache_manager.add(key="a", value=1)
        cache_manager.add(key="b", value=2)
        cache_manager.add(key="c", value=3)
        #
        for k, v in {"a": 1, "b": 2, "c": 3}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata

        # 重新添加 a，并访问 a
        #   优先级变为 a、c、b
        cache_manager.add(key="a", value=3)
        check_consistency(cache_manager.get(key="a"), 3)
        #
        for k, v in {"a": 3, "b": 2, "c": 3}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["d", "e"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)

        # 添加数据 d
        #   优先级变为 d、a、c、b，超过upper_bound，触发重整到refactor_size大小，变为 d、a
        cache_manager.add(key="d", value=4)
        #
        for k, v in {"a": 3, "d": 4}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["b", "c"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)

        # 继续添加 e、f、g
        #    重整后优先级依次为 g、f、e
        cache_manager.add(key="e", value=5)
        cache_manager.add(key="f", value=6)
        cache_manager.add(key="g", value=7)
        #
        for k, v in {"e": 5, "f": 6, "g": 7}.items():
            assert cache_manager.has(key=k)
        for k in ["a", "d"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)


def test_cache_manager_with_fifo_strategy():
    print("test Cache_Manager with FIFO_Strategy")

    strategy = ":by_initial_time:FIFO"
    for cache in [":in_memory:Memo", ]:
        cache_manager = Cache_Manager(upper_bound=3, refactor_size=2, strategy=strategy, cache=cache)

        # 添加数据 a、b、c
        #   优先级依次为 c、b、a
        cache_manager.add(key="a", value=1)
        cache_manager.add(key="b", value=2)
        cache_manager.add(key="c", value=3)
        #
        for k, v in {"a": 1, "b": 2, "c": 3}.items():
            check_consistency(cache_manager.get(key=k), v)

        # 无论后面怎么读取，优先级都不变
        for k, v in {"b": 2, "c": 3, "a": 1}.items():
            for _ in range(random.randint(3, 6)):
                cache_manager.get(key=k)
            cache_manager.add(key=k, value=v)

        # 添加数据 d
        #   优先级变为 d、c、b、a，超过upper_bound，触发重整到refactor_size大小，变为 d、c
        cache_manager.add(key="d", value=4)
        #
        for k, v in {"c": 3, "d": 4}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["b", "a"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)

        # 继续添加 e、f、g
        #    重整后优先级依次为 g、f、e
        cache_manager.add(key="e", value=5)
        cache_manager.add(key="f", value=6)
        cache_manager.add(key="g", value=7)
        #
        for k, v in {"e": 5, "f": 6, "g": 7}.items():
            assert cache_manager.has(key=k)
        for k in ["a", "d"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)


def test_cache_manager_with_lfu_strategy():
    print("test Cache_Manager with LFU_Strategy")

    strategy = ":by_counts:LFU"
    for cache in [":in_memory:Memo", ]:
        cache_manager = Cache_Manager(upper_bound=3, refactor_size=2, strategy=strategy, cache=cache)

        # 添加数据 a、b、c，然后各访问 2、3、4
        #   优先级依次为 c、b、a
        cache_manager.add(key="a", value=1)
        cache_manager.add(key="b", value=2)
        cache_manager.add(key="c", value=3)
        #
        for k, v in {"a": 1, "b": 2, "c": 3}.items():
            for _ in range(v + 1):
                check_consistency(cache_manager.get(key=k), v)
        check_consistency(cache_manager.get(key="d", default_factory=lambda: "fuck"), "fuck")

        # 重新添加 c
        #   优先级变为 b、a、c
        cache_manager.add(key="c", value=4, b_allow_overwrite=True)
        check_consistency(cache_manager.get(key="c"), 4)
        #
        for k, v in {"a": 1, "b": 2, "c": 4}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["d", "e"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)

        # 添加数据 d
        #   优先级变为 b、a、c、d，超过upper_bound，触发重整到refactor_size大小，变为 b、a
        cache_manager.get(key="d", default=4, b_add_if_not_found=True)
        #
        for k, v in {"a": 1, "b": 2}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["c", "d"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)

        # 继续添加 e、f、g，访问一次 e
        #    重整后优先级依次为 b、a、g
        cache_manager.get(key="e", default=5, b_add_if_not_found=True)
        cache_manager.add(key="f", value=6)
        cache_manager.add(key="g", value=7)
        #
        for k, v in {"g": 7, "a": 1, "b": 2}.items():
            assert cache_manager.has(key=k)
        for k in ["f", "e"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)


def test_cache_manager_with_lst_strategy():
    print("test Cache_Manager with LST_Strategy")

    strategy = ":by_survival_time:LST"
    for cache in [":in_memory:Memo", ]:
        cache_manager = Cache_Manager(upper_bound=3, refactor_size=2, strategy=strategy, cache=cache)

        # 添加数据 a、b、c，然后读取2次 b，读取1次 c，读取3次 a
        #   优先级依次为 a、c、b
        cache_manager.add(key="a", value=1)
        cache_manager.add(key="b", value=2)
        cache_manager.add(key="c", value=3)
        #
        for k, v in {"b": 2, "c": 1, "a": 3}.items():
            for _ in range(v):
                time.sleep(0.1)
                cache_manager.get(key=k)

        # 添加数据 d
        #   优先级变为 a、c、b、d，超过upper_bound，触发重整到refactor_size大小，变为 a、c
        cache_manager.get(key="d", default=4, b_add_if_not_found=True)
        #
        for k, v in {"a": 1, "c": 3}.items():
            assert cache_manager.has(key=k)
            check_consistency(cache_manager.cache.read(key=k), v)  # 不影响 metadata
        for k in ["b", "d"]:
            assert not cache_manager.has(key=k)
            with pytest.raises(KeyError):
                cache_manager.get(key=k)



from kevin_toolbox.math.utils import split_integer_most_evenly
from kevin_toolbox.data_flow.file.markdown.table import Table_Format, get_format, padding_misaligned_values


def complete_to_matrix(content_s, orientation="vertical", chunk_nums=None, chunk_size=None):
    """
        生成表格

        参数：
            content_s:              <dict> 内容
                                        目前支持 Table_Format 中的两种输入模式：
                                            1.简易模式：
                                                content_s = {<title>: <list of value>, ...}
                                                此时键作为标题，值作为标题下的一系列值。
                                                由于字典的无序性，此时标题的顺序是不能保证的，若要额外指定顺序，请使用下面的 完整模式。
                                            2. 完整模式:
                                                content_s = {<index>: {"title": <title>,"values":<list of value>}, ...}
                                                此时将取第 <index> 个 "title" 的值来作为第 <index> 个标题的值。values 同理。
                                                该模式允许缺省某些 <index>，此时这些 <index> 对应的行/列将全部置空。
            orientation:            <str> 表格的方向
                                        支持以下值：
                                            "vertical" / "v":       纵向排列，亦即标题在第一行
                                            "horizontal" / "h":     横向排列，亦即标题在第一列
            chunk_nums:             <int> 将表格平均分割为多少份进行并列显示。
            chunk_size:             <int> 将表格按照最大长度进行分割，然后并列显示。
                注意：以上两个参数只能设置一个，同时设置时将报错
    """
    # 检验参数
    assert chunk_nums is None or 1 <= chunk_nums
    assert chunk_size is None or 1 <= chunk_size
    assert orientation in ["vertical", "horizontal", "h", "v"]
    assert get_format(content_s) is Table_Format.COMPLETE_DICT

    # 当不同标题下的 values 的长度不相等时，先使用 padding_misaligned_values() 来进行对齐
    content_s = padding_misaligned_values(content_s=content_s, padding_value="")
    max_length = len(list(content_s.values())[0]["values"])

    # 补充缺省的 title
    for i in range(max(content_s.keys()) + 1):
        if i not in content_s:
            content_s[i] = {"title": "", "values": [""] * max_length}
    # 按照 chunk_nums 或者 chunk_size 对表格进行分割
    if chunk_nums is not None or chunk_size is not None:
        if chunk_nums is not None:
            split_len_ls = split_integer_most_evenly(x=max_length, group_nums=chunk_nums)
        else:
            split_len_ls = [chunk_size] * (max_length // chunk_size)
            if max_length % chunk_size != 0:
                split_len_ls += [max_length % chunk_size]
        max_length = max(split_len_ls)
        temp = dict()
        beg = 0
        for i, new_length in enumerate(split_len_ls):
            end = beg + new_length
            temp.update({k + i * len(content_s): {"title": v["title"],
                                                  "values": v["values"][beg:end] + [""] * (max_length - new_length)} for
                         k, v in content_s.items()})
            beg = end
        content_s = temp

    # 转换
    row_ls = []
    if orientation in ["vertical", "v"]:
        row_ls.append([content_s[i]["title"] for i in range(len(content_s))])
        for row in zip(*[content_s[i]["values"] for i in range(len(content_s))]):
            row_ls.append(row)
    else:
        for i in range(len(content_s)):
            row_ls.append([content_s[i]["title"]] + content_s[i]["values"])

    return dict(matrix=row_ls, orientation=orientation, chunk_size=chunk_size, chunk_nums=chunk_nums,
                b_remove_empty_lines=chunk_size is not None or chunk_nums is not None)


if __name__ == '__main__':
    from kevin_toolbox.data_flow.file.markdown.table import convert_format

    content_s = complete_to_matrix(
        content_s=convert_format(
            content_s={'y/n': ['False', 'False', 'False', 'False', 'False', 'True', 'True', 'True', 'True', 'True'],
                       'a': ['5', '8', '7', '6', '9', '2', '1', '4', '0', '3'],
                       'b': ['', '', '', '', '', '6', '4', ':', '2', '8']},
            output_format=Table_Format.COMPLETE_DICT
        ),
        orientation="v", chunk_size=4
    )


    def _show_table(row_ls):
        """
            生成表格文本

            参数：
                row_ls:                 <list of row>
        """
        table = ""
        for idx, row in enumerate(row_ls):
            row = [f'{i}' for i in row]
            table += "| " + " | ".join(row) + " |\n"
            if idx == 0:
                table += "| " + " | ".join(["---"] * len(row)) + " |\n"
        return table


    doc = _show_table(content_s["matrix"])
    print(doc)



from abc import ABC, abstractmethod


class Strategy_Base(ABC):
    """
        缓存管理策略的基类

        要求实现：
            - notified_by_write_of_cache()
            - notified_by_read_of_cache()
            - notified_by_remove_of_cache()
            - notified_by_clear_of_cache()
            - 返回需要删除的条目的键的列表 suggest()
            - 清空所有内容 clear()
        等方法。

        其中 notified_by_xxx_of_cache 系列方法用于接受在 Cache_Manager 中的通讯，比如 notified_by_write_of_cache 方法
            将在 Cache_Manager 中每次调用 cache 的 write 方法时被调用。利用这些通讯接口，使得策略管理器可以提前追踪、感知条目的变化，
            从而进行一些预处理操作，以免在后续需要其提出重构建议时，才当场进行过大的计算量。
    """

    @abstractmethod
    def notified_by_write_of_cache(self, key, value, metadata):
        pass

    @abstractmethod
    def notified_by_read_of_cache(self, key, value, metadata):
        pass

    @abstractmethod
    def notified_by_remove_of_cache(self, key, metadata):
        pass

    @abstractmethod
    def notified_by_clear_of_cache(self):
        self.clear()

    @abstractmethod
    def suggest(self, refactor_size) -> list:
        """返回需要删除的条目的键的列表"""
        pass

    @abstractmethod
    def clear(self):
        pass



from kevin_toolbox.computer_science.algorithm.cache_manager.cache import Cache_Base
from kevin_toolbox.computer_science.algorithm.cache_manager.variable import CACHE_BUILDER_REGISTRY


@CACHE_BUILDER_REGISTRY.register()
class Memo_Cache(Cache_Base):
    """
        基于内存的缓存结构
    """
    name = ":in_memory:Memo_Cache"

    def __init__(self, **kwargs):
        self.cache_s = dict()

    def _read_freely(self, key):
        return self.cache_s[key]

    def _write_freely(self, key, value):
        self.cache_s[key] = value

    def _remove_freely(self, key):
        self.cache_s.pop(key)

    def has(self, key):
        return key in self.cache_s

    def len(self):
        return len(self.cache_s)

    def clear(self):
        self.cache_s.clear()


# 添加其他别名
for name in [":in_memory:Memo", ":in_memory:MC", ":in_memory:memory"]:
    CACHE_BUILDER_REGISTRY.add(obj=Memo_Cache, name=name, b_force=False, b_execute_now=False)



import re


def find_tables(text, b_compact_format=True):
    """
        查找文本中的表格

        参数:
            text:               <str> 文本
            b_compact_format:   <bool> 是否只返回 table 部分
                                    默认为 True，此时返回 table_ls，其中每个元素是一个 MATRIX 格式的表格
                                    当设置为 False，此时返回 (table_ls, part_slices_ls, table_idx_ls)，
                                        其中 part_slices_ls 是表格和表格前后文本在 text 中对应的 slice，
                                        而 table_idx_ls 指出了 part_slices_ls 中第几个元素对应的是表格，
                                        table_idx_ls 与 table_ls 依次对应。
    """
    text = "\n\n" + text + "\n\n"  # 前后使用哨兵包围
    matches = re.finditer(r'\n{2,}', text, re.DOTALL)

    table_ls = []
    part_slices_ls = []
    table_idx_ls = []
    #
    match = next(matches)
    start, sub_start = match.start(), match.end()
    assert sub_start - start >= 2
    if sub_start - start > 2:
        part_slices_ls.append([start + 2, sub_start])
    start = sub_start
    #
    for match in matches:
        sub_text = text[sub_start:match.start()]
        ret = _find_table(text=sub_text)
        if ret is not None:
            if start != sub_start:
                part_slices_ls.append([start, sub_start])
            table_idx_ls.append(len(part_slices_ls))
            table_ls.append(ret)
            part_slices_ls.append([sub_start, match.start()])
            start = match.start()
        sub_start = match.end()
    #
    assert sub_start - start >= 2
    if sub_start - start > 2:
        part_slices_ls.append([start, sub_start - 2])
    # 移除前面哨兵
    part_slices_ls = [[i - 2, j - 2] for i, j in part_slices_ls]

    if b_compact_format:
        return table_ls
    else:
        return table_ls, part_slices_ls, table_idx_ls


def _find_table(text):
    # 正则表达式匹配Markdown表格
    table_pattern = re.compile(r'\|([^\n]+)\|', re.DOTALL)
    table_matches = table_pattern.findall(text)
    if len(table_matches) < 2:
        # 因为一个合法的 markdown 表格需要含有表头的分隔线，所以行数至少应该为 2
        return None

    # 去除表头的分隔线
    table_matches.pop(1)
    #
    tables = []  # 每个元素为一行
    for match in table_matches:
        # 分割每一行
        tables.append([i.strip() for i in match.split('|', -1)])

    return {"matrix": tables, "orientation": None}


if __name__ == '__main__':
    # # 示例Markdown表格文本
    # file_path = ""
    # with open(file_path, 'r') as f:
    #     markdown_text = f.read()

    markdown_text = """
| Name | Age | Occupation |
|------|-----|------------|
| Alice | 28  | Engineer   |
| Bob   | 23  | Teacher    |
| Name | Age | Occupation |
| Carol | 32  | Hacker   |
| David | 18  | Student   |

2333

|  | a | b |  | a | b |  | a | b |
| --- | --- | --- | --- | --- | --- | --- | --- | --- |
|  | 0 | 2 |  | 4 | 6 |  | 7 | 9 |
|  | 1 | 3 |  | 5 | 7 |  | 8 | : |
|  | 2 | 4 |  | 6 | 8 |  | 9 | ; |
|  | 3 | 5 |  |  |  |  |  |  |
"""

    # 调用函数并打印结果
    tables = find_tables(text=markdown_text)
    print(tables[0])
    print(tables[1])

    #
    table_ls_, part_slices_ls_, table_idx_ls_ = find_tables(text=markdown_text, b_compact_format=False)
    print(table_idx_ls_)

    for part_slices in  part_slices_ls_:
        print(part_slices)
        print(markdown_text[part_slices[0]:part_slices[1]])




from .remove import remove
from .copy import copy
from .pack import pack
from .unpack import unpack
from .walk import walk, Path_Ignorer, Ignore_Scope
from .find_files_in_dir import find_files_in_dir
from . import path



import os
import matplotlib.pyplot as plt
from kevin_toolbox.patches.for_matplotlib.color import generate_color_list
from kevin_toolbox.patches.for_os.path import replace_illegal_chars


def plot_scatters(data_s, title, x_name, y_name, cate_name=None, output_dir=None, **kwargs):
    paras = {
        "dpi": 200,
        "scatter_size": 5
    }
    paras.update(kwargs)

    plt.clf()
    #
    color_s = None
    if cate_name is not None:
        cates = list(set(data_s[cate_name]))
        color_s = {i: j for i, j in zip(cates, generate_color_list(nums=len(cates)))}
        c = [color_s[i] for i in data_s[cate_name]]
    else:
        c = "blue"
    # 创建散点图
    plt.scatter(data_s[x_name], data_s[y_name], s=paras["scatter_size"], c=c, alpha=0.8)
    #
    plt.xlabel(f'{x_name}')
    plt.ylabel(f'{y_name}')
    plt.title(f'{title}')
    # 添加图例
    if cate_name is not None:
        plt.legend(handles=[
            plt.Line2D([0], [0], marker='o', color='w', label=i, markerfacecolor=j,
                       markersize=min(paras["scatter_size"], 5)) for i, j in color_s.items()
        ])

    if output_dir is None:
        plt.show()
        return None
    else:
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, f'{replace_illegal_chars(title)}.png')
        plt.savefig(output_path, dpi=paras["dpi"])
        return output_path


if __name__ == '__main__':
    data_s_ = dict(
        x=[1, 2, 3, 4, 5],
        y=[2, 4, 6, 8, 10],
        categories=['A', 'B', 'A', 'B', 'A']
    )

    plot_scatters(data_s=data_s_, title='test', x_name='x', y_name='y', cate_name='categories')



from kevin_toolbox.data_flow.file.markdown.table import convert_format, Table_Format, padding_misaligned_values
from kevin_toolbox.data_flow.file.markdown.table.convert import complete_to_matrix


def generate_table(content_s, orientation="vertical", chunk_nums=None, chunk_size=None, b_allow_misaligned_values=False,
                   f_gen_order_of_values=None):
    """
        生成表格

        参数：
            content_s:              <dict> 内容
                                        目前支持 Table_Format 中的两种输入模式：
                                            1.简易模式：
                                                content_s = {<title>: <list of value>, ...}
                                                此时键作为标题，值作为标题下的一系列值。
                                                由于字典的无序性，此时标题的顺序是不能保证的，若要额外指定顺序，请使用下面的 完整模式。
                                            2. 完整模式:
                                                content_s = {<index>: {"title": <title>,"values":<list of value>}, ...}
                                                此时将取第 <index> 个 "title" 的值来作为第 <index> 个标题的值。values 同理。
                                                该模式允许缺省某些 <index>，此时这些 <index> 对应的行/列将全部置空。
            orientation:            <str> 表格的方向
                                        支持以下值：
                                            "vertical" / "v":       纵向排列，亦即标题在第一行
                                            "horizontal" / "h":     横向排列，亦即标题在第一列
            chunk_nums:             <int> 将表格平均分割为多少份进行并列显示。
            chunk_size:             <int> 将表格按照最大长度进行分割，然后并列显示。
                注意：以上两个参数只能设置一个，同时设置时将报错
            b_allow_misaligned_values:  <boolean> 允许不对齐的 values
                                        默认为 False，此时当不同标题下的 values 的长度不相等时，将会直接报错。
                                        当设置为 True 时，对于短于最大长度的 values 将直接补充 ""。
            f_gen_order_of_values:  <callable> 生成values排序顺序的函数
                                        该函数需要接受一个形如 {<title>: <value>, ...} 的 <dict>，并返回一个用于排序的 int/float/tuple
    """
    # 检验参数
    assert chunk_nums is None or 1 <= chunk_nums
    assert chunk_size is None or 1 <= chunk_size
    assert orientation in ["vertical", "horizontal", "h", "v"]
    assert isinstance(content_s, (dict,))

    # 首先转换为完整模式
    content_s = convert_format(content_s=content_s, output_format=Table_Format.COMPLETE_DICT)
    # 对齐 values
    len_ls = [len(v["values"]) for v in content_s.values()]
    max_length = max(len_ls)
    if min(len_ls) != max_length:
        assert b_allow_misaligned_values, \
            f'The lengths of the values under each title are not aligned. ' \
            f'The maximum length is {max_length}, but each length is {len_ls}'
        content_s = padding_misaligned_values(content_s=content_s, padding_value="")
    # 对值进行排序
    if callable(f_gen_order_of_values):
        # 检查是否有重复的 title
        temp = [v["title"] for v in content_s.values()]
        assert len(set(temp)) == len(temp), \
            f'table has duplicate titles, thus cannot be sorted using f_gen_order_of_values'
        idx_ls = list(range(max_length))
        idx_ls.sort(key=lambda x: f_gen_order_of_values({v["title"]: v["values"][x] for v in content_s.values()}))
        for v in content_s.values():
            v["values"] = [v["values"][i] for i in idx_ls]

    # 转换为 matrix 格式
    content_s = complete_to_matrix(content_s=content_s, orientation=orientation, chunk_size=chunk_size,
                                   chunk_nums=chunk_nums)
    # 构建表格
    table = ""
    for idx, row in enumerate(content_s["matrix"]):
        row = [f'{i}' for i in row]
        table += "| " + " | ".join(row) + " |\n"
        if idx == 0:
            table += "| " + " | ".join(["---"] * len(row)) + " |\n"
    return table


if __name__ == '__main__':
    # content_s = {0: dict(title="a", values=[1, 2, 3]), 2: dict(title="b", values=[4, 5, 6])}
    # doc = generate_table(content_s=content_s, orientation="h")
    # print(doc)

    # from collections import OrderedDict
    #
    # content_s = OrderedDict({
    #     "y/n": [True] * 5 + [False] * 5,
    #     "a": list(range(10)),
    #     "b": [chr(i) for i in range(50, 60, 2)]
    # })
    # doc = generate_table(content_s=content_s, orientation="v", chunk_size=4, b_allow_misaligned_values=True,
    #                      f_gen_order_of_values=lambda x: (-int(x["y/n"] is False), -(x["a"] % 3)))
    # print(doc)
    # import os
    #
    # with open(os.path.join(
    #         "/home/SENSETIME/xukaiming/Desktop/my_repos/python_projects/kevin_toolbox/kevin_toolbox/data_flow/file/markdown/test/test_data/for_generate_table",
    #         f"data_5.md"), "w") as f:
    #     f.write(doc)

    doc = generate_table(
        content_s={'y/n': ['False', 'False', 'False', 'False', 'False', 'True', 'True', 'True', 'True', 'True'],
                   'a': ['5', '8', '7', '6', '9', '2', '1', '4', '0', '3'],
                   'b': ['', '', '', '', '', '6', '4', ':', '2', '8']},
        orientation="v", chunk_size=4, b_allow_misaligned_values=True,
        f_gen_order_of_values=lambda x: (-int(eval(x["y/n"]) is False), -(int(x["a"]) % 3))
    )
    print(doc)



import os
import shutil
from kevin_toolbox.patches.for_os import remove


def copy(src, dst, follow_symlinks=True, remove_dst_if_exists=False):
    """
        复制文件/文件夹/软连接

        参数：
            follow_symlinks:        <boolean> 是否跟随符号链接
                                    对于 src 是软连接或者 src 指向的目录下具有软连接的情况，
                                        当设置为 True 时，会复制所有链接指向的实际文件或目录内容。
                                        当设置为 False 时，则仅仅软连接本身，而不是指向的内容
                                    默认为 True
            remove_dst_if_exists:   <boolean> 当目标存在时，是否尝试进行移除
    """
    assert os.path.exists(src), f'failed to copy, src not exists: {src}'
    if remove_dst_if_exists:
        remove(path=dst, ignore_errors=True)
    assert not os.path.exists(dst), f'failed to copy, dst exists: {dst}'

    os.makedirs(os.path.dirname(dst), exist_ok=True)
    if os.path.isdir(src):
        if not follow_symlinks and os.path.islink(src):
            # 如果是符号链接，并且我们跟随符号链接，则复制链接本身
            os.symlink(os.readlink(src), dst)
        else:
            # 否则，递归复制目录
            shutil.copytree(src, dst, symlinks=not follow_symlinks)
    else:
        # 复制文件
        shutil.copy2(src, dst, follow_symlinks=follow_symlinks)



import os
import re

ILLEGAL_CHARS = {
    '<': '＜',
    '>': '＞',
    ':': '：',
    '"': '＂',
    '/': '／',
    '\\': '＼',
    '|': '｜',
    '?': '？',
    '*': '＊'
}

ILLEGAL_CHARS_PATTERN = re.compile(r'[<>:"/\\|?*]')


def find_illegal_chars(file_name, b_is_path=False):
    """
        找出给定的文件名/路径中出现了哪些非法符号
            所谓非法符号是指在特定系统（win/mac等）中文件名不允许出现的字符，
            不建议在任何系统下使用带有这些符号的文件名，即使这些符号在当前系统中是合法的，
            以免在跨系统时出现兼容性问题

        参数：
            file_name:      <str>
            b_is_path:      <bool> 是否将file_name视为路径
                                默认为 False
                                当设置为 True 时，将会首先将file_name分割，再逐级从目录名中查找
    """
    global ILLEGAL_CHARS_PATTERN

    if b_is_path:
        temp = [i for i in file_name.split(os.sep, -1) if len(i) > 0]
    else:
        temp = [file_name]
    res = []
    for i in temp:
        res.extend(ILLEGAL_CHARS_PATTERN.findall(i))
    return res


if __name__ == '__main__':
    file_path = '//data0//b/<?>.md'
    print(find_illegal_chars(file_name=file_path, b_is_path=True))
    print(find_illegal_chars(file_name=file_path, b_is_path=False))



import os
import numpy as np
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import seaborn as sns
from kevin_toolbox.patches.for_os.path import replace_illegal_chars


def plot_confusion_matrix(data_s, title, gt_name, pd_name, label_to_value_s=None, output_dir=None,
                          replace_zero_division_with=0, **kwargs):
    """
        计算并绘制混淆矩阵

        参数：
            replace_zero_division_with:         <float> 对于在normalize时引发除0错误的矩阵元素，使用何种值进行替代
                                                    建议使用 np.nan 或者 0
    """
    paras = {
        "dpi": 200,
        "normalize": None,  # "true", "pred", "all",
        "b_return_cfm": False,  # 是否输出混淆矩阵
    }
    paras.update(kwargs)

    value_set = set(data_s[gt_name]).union(set(data_s[pd_name]))
    if label_to_value_s is None:
        label_to_value_s = {f'{i}': i for i in value_set}
    else:
        # assert all(i in value_set for i in label_to_value_s.values())
        pass
    # 计算混淆矩阵
    cfm = confusion_matrix(y_true=data_s[gt_name], y_pred=data_s[pd_name], labels=list(label_to_value_s.values()),
                           normalize=paras["normalize"])
    # replace with nan
    if paras["normalize"] is not None:
        if paras["normalize"] == "all":
            if cfm.sum() == 0:
                cfm[cfm == 0] = replace_zero_division_with
        else:
            check_axis = 1 if paras["normalize"] == "true" else 0
            temp = np.sum(cfm, axis=check_axis, keepdims=False)
            for i in range(len(temp)):
                if temp[i] == 0:
                    if check_axis == 0:
                        cfm[:, i] = replace_zero_division_with
                    else:
                        cfm[i, :] = replace_zero_division_with

    # 绘制混淆矩阵热力图
    plt.clf()
    plt.figure(figsize=(8, 6))
    sns.heatmap(cfm, annot=True, fmt='.2%' if paras["normalize"] is not None else 'd',
                xticklabels=list(label_to_value_s.keys()), yticklabels=list(label_to_value_s.keys()),
                cmap='viridis')

    plt.xlabel(f'{pd_name}')
    plt.ylabel(f'{gt_name}')
    plt.title(f'{title}')

    if output_dir is None:
        plt.show()
        output_path = None
    else:
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, f'{replace_illegal_chars(title)}.png')
        plt.savefig(output_path, dpi=paras["dpi"])

    if paras["b_return_cfm"]:
        return output_path, cfm
    else:
        return output_path


if __name__ == '__main__':
    # 示例真实标签和预测标签
    y_true = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 5])
    y_pred = np.array([0, 2, 1, 0, 2, 1, 0, 1, 1, 5])

    plot_confusion_matrix(data_s={'a': y_true, 'b': y_pred},
                          title='test', gt_name='a', pd_name='b',
                          label_to_value_s={"A": 5, "B": 0, "C": 1, "D": 2, "E": 3},
                          # output_dir=os.path.join(os.path.dirname(__file__), "temp"),
                          replace_zero_division_with=-1,
                          normalize="all")



from kevin_toolbox.patches.for_numpy.random import get_rng


def set_rng_state(state, rng=None):
    """
        加载状态到随机生成器中

        参数：
            state:
            rng:            当设置为 None 时，将通过 get_rng() 新建一个 rng，然后加载状态并返回
    """
    rng = rng or get_rng(seed=114514)
    rng.bit_generator.state = state
    return rng



import re
import os
import inspect
import pkgutil
import weakref
import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.patches import for_os
from kevin_toolbox.patches.for_os import Path_Ignorer, Ignore_Scope


class Registry:
    """
        注册器
            具有以下功能
            - 管理成员，包括添加 add()、获取 get() pop() 成员等
            - 支持通过装饰器 register() 来添加成员
            - 支持通过 collect_from_paths() 搜索指定的路径，当该路径下的模块被 register() 装饰器包裹或者通过 add() 添加成员时，将自动导入
            （用于解决python中的模块是惰性的问题）

        使用方法：
            以目录结构：
                xx/——— modules_dir
                 |       |——— a.py
                 |       |___ ...
                 |——— test.py
            为例。
            我们的目标是在 test.py 中，构建一个能够自动加载 modules_dir 中待注册成员的 Registry 实例。
            具体步骤如下：
                1. 在文件 test.py 中：
                    # 创建注册器实例
                    DB = Registry(uid="DB")
                    # 设置搜索路径
                    DB.collect_from_paths(path_ls=["xx/modules_dir"])

                2. 在 modules_dir 下需要加载的成员中，以 a.py 为例：
                    # 导入注册器实例
                    from xx.test import DB
                    # 使用 DB.register() 装饰器注册
                    @DB.register(name=":my_module")
                    class A:
                        ...
                    # 使用 add() 函数注册
                    DB.add(obj=233, name=":var:c")

                3. 在文件 test.py 中：
                    # 获取已注册成员
                    module = DB.get(name=":my_module")

                4. 如果需要在其他文件中获取已有注册器实例，除了用 import 之外，还可以直接用 uid
                    # 获取指定 uid 下的实例
                    temp = Registry(uid="DB")
    """
    name = "registry made by kevin"
    __instances = weakref.WeakValueDictionary()  # {uid: instance_i, ...} 用于保存创建的实例
    __counter = 0

    def __new__(cls, *args, **kwargs):
        """
            __new__函数返回的实例，将作为self参数被传入到__init__函数。
                如果__new__函数返回一个已经存在的实例（不论是哪个类的），__init__还是会被调用的，所以要特别注意__init__中对变量的赋值。
        """
        if "uid" in kwargs:
            uid = kwargs["uid"]
        else:
            while cls.__counter in cls.__instances:
                cls.__counter += 1
            uid = cls.__counter

        # 获取实例
        if uid in cls.__instances:
            # 返回已存在的实例
            self = cls.__instances[uid]
        else:
            # 传入 __init__ 中新建一个实例
            self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        """
            参数：
                uid:                <hashable> 实例的唯一标识符
                                        - 不设置。新建一个实例，并自动分配一个整数作为 uid（一般是此时已有实例的数量-1），
                                            并将该实例记录到类变量 instances_of_Registry 中。
                                        - 设置为 None。新建一个实例，但不记录到类变量 instances_of_Registry 中。
                                        - 设置为其他值。根据 uid 到类变量 instances_of_Registry 的已有实例中寻找相同的实例，
                                            若命中则返回已有实例，若无则以该 uid 新建一个实例并添加到 instances_of_Registry 中。
                                            （使用该特性可以构造单例模式）
                                        默认为不设置。
        """
        try:
            getattr(self, "uid")
        except:
            pass
        else:
            return  # 如果是从 __new__ 中获取的已有的实例则不重复进行参数赋值
        self.database = dict()
        #
        self.uid = kwargs.get("uid", Registry.__counter)
        self._path_to_collect = []
        self._item_to_add = []
        # 记录到 __instances 中
        if self.uid is not None:
            Registry.__instances[self.uid] = self

    def add(self, obj, name=None, b_force=False, b_execute_now=True):
        """
            注册

            参数：
                obj：            待注册成员
                                    可以是函数、类或者callable的实例
                                    也可以是各种 int、float、str变量
                                    总之一切对象皆可
                name：           <str> 成员名称
                                    默认为 None，此时将从被注册对象 obj 的属性中尝试推断出其名称。
                                        若 obj 中有 name 或者 __name__ 属性（优先选择name），则推断出的名称是 f'{obj.name}' 或者 f':{obj.__name__}'；
                                            进一步若有 version 属性，则为 f'{obj.name}:{obj.version}'
                                        否则报错。
                                        比如下面的类：
                                            class A:
                                                version="1.0"
                                        的默认注册名称将是 ":A:1.0"
                                        另一种等效的写法是：
                                            class A:
                                                name=":A:1.0"
                                    对于 int、str 和其他没有 name 或者 __name__ 属性的变量则必须要手动指定 name 参数。
                        需要注意的是，成员的名称确定了其在注册器内部 database 中的位置，名称的解释方式参考 get_value() 中的介绍。
                        因此不同的名称可能指向了同一个位置。
                b_force：          <boolean> 是否强制注册
                                    默认为 False，此时当 name 指向的位置上已经有成员或者需要强制修改database结构时，将不进行覆盖而直接跳过，注册失败
                                    当设置为 True，将会强制覆盖
                b_execute_now:      <boolean> 现在就执行注册
                                        默认为 True，否则将等到第一次执行 get() 函数时才会真正尝试注册

            返回：
                <boolean>   是否注册成功
        """
        # 检验参数
        if name is None:
            name = getattr(obj, "name", None)
            if name is not None:
                name = f'{name}'
            else:
                name = getattr(obj, "__name__", None)
                name = f':{name}' if name is not None else name
            if name is not None:
                version = getattr(obj, "version", None)
                if version is not None:
                    name += f':{version}'
        assert isinstance(name, (str,))

        #
        if not b_execute_now:
            self._item_to_add.append(dict(obj=obj, name=name, b_force=b_force, b_execute_now=True))
            return True

        # 尝试注册
        temp = ndl.set_value(var=ndl.copy_(var=self.database, b_deepcopy=False), name=name, value=obj,
                             b_force=True)
        # check
        if not b_force:
            inc_node_nums = ndl.count_leaf_node_nums(var=obj) if isinstance(obj, (list, dict)) else 1  # 应该增加的节点数量
            if ndl.count_leaf_node_nums(var=temp) != ndl.count_leaf_node_nums(var=self.database) + inc_node_nums:
                # print(f'registration failed, name {name} may be a conflict with an '
                #       f'existing member in {[i for i, j in ndl.get_nodes(var=self.database)]}')
                return False

        self.database = temp
        return True

    def get(self, name, b_pop=False, **kwargs):
        """
            获取

            参数：
                name:           <str> 成员名称
                b_pop:          <boolean> 取值的同时移除该成员
                                    默认为 False
                default:        默认值
                                    找不到时，若无默认值则报错，否则将返回默认值
        """
        # 加载待注册成员
        while self._item_to_add or self._path_to_collect:
            if len(self._item_to_add) > 0:
                for i in self._item_to_add:
                    self.add(**i)
                self._item_to_add.clear()
            if len(self._path_to_collect) > 0:
                for i in self._path_to_collect:
                    i.setdefault("caller_file", inspect.stack()[1].filename)
                    self.collect_from_paths(**i)
                self._path_to_collect.clear()

        return ndl.get_value(var=self.database, name=name, b_pop=b_pop, **kwargs)

    def clear(self):
        self.database.clear()

    # -------------------- 装饰器 --------------------- #

    def register(self, name=None, b_force=False, b_execute_now=True):
        """
            用于注册成员的装饰器
                成员可以是函数、类或者callable的实例

            参数：
                name：           <str> 成员名称
                                    默认为 None
                b_force：        <boolean> 是否强制注册
                                    默认为 False
                b_execute_now:  <boolean> 现在就执行注册
                                    默认为 True
                                （以上参数具体参考 add() 函数介绍）
        """

        def wrapper(obj):
            nonlocal self, name, b_force, b_execute_now
            self.add(obj, name=name, b_force=b_force, b_execute_now=b_execute_now)
            return obj

        return wrapper

    # -------------------- 通过路径添加 --------------------- #

    def collect_from_paths(self, path_ls=None, ignore_s=None, b_execute_now=False, **kwargs):
        """
            遍历 path_ls 下的所有模块，并自动导入其中主要被注册的部分
                比如被 register() 装饰器包裹或者通过 add() 添加的部分

            参数：
                path_ls:            <list of paths> 需要搜索的目录
                ignore_s:           <list/tuple of dict> 在搜索遍历目录时候要执行的排除规则
                                        具体设置方式参考 patches.for_os.walk() 中的 ignore_s 参数。
                                        比如使用下面的规则就可以排除待搜索目录下的所有 temp/ 和 test/ 文件夹：
                                        [
                                            {
                                                "func": lambda _, __, path: os.path.basename(path) in ["temp", "test"],
                                                "scope": ["root", "dirs"]
                                            },
                                        ]
                b_execute_now:      <boolean> 现在就执行导入
                                        默认为 False，将等到第一次执行 get() 函数时才会真正尝试导入

                        注意，在某个文件中使用 Registry 实例时，应尽量避免下面的情况：
                            1. 在当前脚本中显式导入该实例前，调用了其他脚本执行了该实例的 collect_from_paths() 函数，且设置 b_execute_now=True，
                                此时若导入的成员中有类，且该类继承自某个父类，且在初始化时使用了 super(xx,self).__init__ 继承初始化函数，将出现
                                TypeError: super(type, obj): obj must be an instance or subtype of type 的错误
                            2. 在模块的 __init__.py 文件中使用 collect_from_paths() 或间接通过 get() 调用 collect_from_paths()
                            3. collect_from_paths() 函数中的搜索路径中包含了调用该函数的文件位置。
                        为了避免情况 1，应该尽量避免设置 b_execute_now=True。
                            或者省略 super(xx,self).__init__ 中的参数改为 super().__init__
        """
        # 检查调用位置
        caller_file = kwargs.get("caller_file", inspect.stack()[1].filename)
        assert os.path.basename(caller_file) != "__init__.py", \
            f'calling Registry.collect_from_paths() in __init__.py is forbidden, file: {caller_file}.\n' \
            f'you can call it in other files, and then import the result of the call in __init__.py'

        # 根据 ignore_s 构建 Path_Ignorer
        path_ignorer = ignore_s if isinstance(ignore_s, (Path_Ignorer,)) else Path_Ignorer(ignore_s=ignore_s)

        #
        if not b_execute_now:
            self._path_to_collect.append(
                dict(path_ls=path_ls, ignore_s=path_ignorer, b_execute_now=True, caller_file=caller_file))
            return

        #
        if path_ls is not None:
            temp = []
            for path in filter(lambda x: os.path.isdir(x), path_ls):
                temp.append(path)
                for root, dirs, _ in for_os.walk(path, topdown=False, ignore_s=path_ignorer, followlinks=True):
                    temp.extend([os.path.join(root, i) for i in dirs])
            # 从深到浅导入，可以避免继承引起的 TypeError: super(type, obj) 类型错误
            path_ls = list(set(temp))
            path_ls.sort(reverse=True)
        path_set = set(path_ls)

        temp = None
        for loader, module_name, is_pkg in pkgutil.walk_packages(path_ls):
            # 判断该模块是否需要导入
            #   （快速判断）判断该模块所在目录是否在 path_set 中
            if loader.path not in path_set:
                continue

            # 使用 find_spec 代替 find_module
            module_spec = loader.find_spec(module_name)
            if module_spec is None:
                continue
            path = module_spec.origin  # 获取模块的路径

            # is_pkg:
            #   - 为 True 时表示当前遍历到的模块是一个包（即一个包含其他模块或子包的目录）
            #   - 为 False 时表示当前模块是一个普通的 Python 模块（文件），不包含其他模块或子包。
            if is_pkg:
                #   若是目录形式的 package，判断是否满足 Path_Ignorer 中的 dirs 对应的规则
                # path = os.path.dirname(loader.find_module(module_name).path)
                path = os.path.dirname(path)
                if path_ignorer(Ignore_Scope.DIRS, True, os.path.islink(path), path):
                    continue
            else:
                #   若该模块是 module，判断该模块的文件路径是否满足 Path_Ignorer 中的 files 对应的规则
                # path = loader.find_module(module_name).path
                if path_ignorer(Ignore_Scope.FILES, False, os.path.islink(path), path):
                    continue
                #   若该模块与调用的文件相同，则报错。
                if path == caller_file:
                    # collect_from_paths() 函数中的搜索路径不应该包含调用该函数文件。
                    #   因为这样将会导致该函数被自己无限递归调用。
                    # 要避免这样的错误，你可以选择：
                    #   1. 将该函数的调用位置放置在待搜索路径外；
                    #   2. 使用 ignore_s 参数来避免加载该函数的调用位置。
                    raise RuntimeError(
                        f'Registry.collect_from_paths(): \n'
                        f'\tThe search path in a function should not include the file location from which it is called. \n'
                        f'\tBecause this will cause the function to be called infinitely recursively by itself. \n'
                        f'To avoid such errors, you can choose: \n'
                        f'\t1. Place the calling location of this function ({path}) outside the path to be searched; \n'
                        f'\t2. Use the ignore_s parameter to avoid searching the calling location of the function, '
                        f'such as {{"func": lambda _, __, path: path == "{path}", "scope": ["files",]}}'
                    )
            # 加载模块
            # module = loader.find_module(module_name).load_module(module_name)
            module = module_spec.loader.load_module(module_name)

            # 选择遍历过程中第一次找到的 Registry 实例
            if temp is None:
                for name, obj in inspect.getmembers(module):
                    if getattr(obj, "name", None) == Registry.name and getattr(obj, "uid", None) == self.uid:
                        temp = obj
                        break

        if temp is not None:
            self.database = temp.database

    # -------------------- 其他 --------------------- #

    @property
    def instances_of_Registry(self):
        return dict(Registry.__instances)


UNIFIED_REGISTRY = Registry(uid="UNIFIED_REGISTRY")



import pytest
from kevin_toolbox.patches.for_test import check_consistency

import os
import numpy as np

from kevin_toolbox.data_flow.file import kevin_notation
from kevin_toolbox.data_flow.file.kevin_notation.test.test_data.data_all import metadata_ls, content_ls, file_path_ls


@pytest.mark.parametrize("expected_metadata, expected_content, file_path",
                         zip(metadata_ls, content_ls, file_path_ls))
def test_write(expected_metadata, expected_content, file_path):
    print("test write()")

    """
    当写入的列的元素不一致时，是否能正常报错
    """

    # 新建
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data/temp", os.path.basename(file_path))

    # 字典方式写入
    if len(expected_content) > 1:
        with pytest.raises(AssertionError):
            list(expected_content.values())[0].clear()
            kevin_notation.write(metadata=expected_metadata, content=expected_content, file_path=file_path)



from kevin_toolbox.data_flow.file.markdown.table import get_format, Table_Format


def padding_misaligned_values(content_s, padding_value=""):
    """
        将标题下长度不相等的 values 补齐
    """
    format_ = get_format(content_s)
    if format_ is Table_Format.COMPLETE_DICT:
        v_ls = [v["values"] for v in content_s.values()]
    elif format_ is Table_Format.SIMPLE_DICT:
        v_ls = list(content_s.values())
    else:
        raise ValueError(f"unsupported format {format_}")

    len_ls = [len(v) for v in v_ls]
    max_length = max(len_ls)
    if min(len_ls) != max_length:
        for v in v_ls:
            v.extend([padding_value] * (max_length - len(v)))

    return content_s



import torch
import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.computer_science.algorithm.statistician import Accumulator_Base


class Accumulator_for_Ndl:
    """
        适用于 ndl 结构的统计器
    """

    def __init__(self, accumulator_builder):
        """
            参数：
                accumulator_builder:        ndl叶节点统计器的构造函数
        """
        assert callable(accumulator_builder) or isinstance(accumulator_builder, Accumulator_Base)
        self.accumulator_builder = accumulator_builder

        self.var = None

    def add(self, var, **kwargs):
        if self.var is None and isinstance(var, (dict, list)):
            self.var = type(var)()
        for name, value in ndl.get_nodes(var=var, level=-1, b_strict=True):
            accumulator = ndl.get_value(var=self.var, name=name, default=None)
            if accumulator is None:
                accumulator = self.accumulator_builder()
                self.var = ndl.set_value(var=self.var, name=name, value=accumulator, b_force=True)
            value = value.detach().cpu().numpy() if torch.is_tensor(value) else value
            accumulator.add(value, **kwargs)

    def add_sequence(self, var_ls, **kwargs):
        for var in var_ls:
            self.add(var, **kwargs)

    def get(self, **kwargs):
        return ndl.traverse(
            var=ndl.copy_(var=self.var, b_deepcopy=False),
            match_cond=lambda _, __, v: not isinstance(v, (dict, list)) and hasattr(v, "get"), action_mode="replace",
            converter=lambda _, v: v.get(**kwargs)
        )


if __name__ == '__main__':
    from kevin_toolbox.data_flow.file import markdown
    import numpy as np
    from kevin_toolbox.computer_science.algorithm.statistician import Average_Accumulator

    worker = Accumulator_for_Ndl(accumulator_builder=Average_Accumulator)

    worker.add({
        1: 2.1,
        "233": torch.ones(10),
        "543": [
            np.array([1, 2, 3]),
            np.array([4, 5, 6]),
        ]
    }, weight=0.8)

    worker.add({
        1: 3.1,
        "233": torch.zeros(10),
        "543": [
            np.array([0, 2, 3]),
            np.array([0, 5, 6]),
        ]
    }, weight=1.4)

    print(markdown.generate_list(var=worker.get()))



from enum import Enum


class Table_Format(Enum):
    """
        表格的几种模式
            1.simple_dict 简易字典模式：
                content_s = {<title>: <list of value>, ...}
                此时键作为标题，值作为标题下的一系列值。
                由于字典的无序性，此时标题的顺序是不能保证的，若要额外指定顺序，请使用下面的 完整模式。
            2. complete_dict 完整字典模式:
                content_s = {<index>: {"title": <title>, "values": <list of value>}, ...}
                此时将取第 <index> 个 "title" 的值来作为第 <index> 个标题的值。values 同理。
                该模式允许缺省某些 <index>，此时这些 <index> 对应的行/列将全部置空。
            3. matrix 矩阵形式：
                content_s = {"matrix": [[...], [...], ...], "orientation":...(, "chunk_nums":..., "chunk_size":...)}
                其中，必要的键值对有：
                    "matrix":           以 list of row 形式保存表格的内容
                    "orientation":      指定表格的解释方向
                                            当为 "vertical" 或 "v" 时，表格为竖直方向，此时第一行为标题，
                                            为 "horizontal" 或 "h" 时，表格为水平方向，此时第一列为标题
                可选键值对有：
                    "chunk_nums":       表格是平均分割为多少份进行并列显示。
                    "chunk_size":       表格是按照最大长度进行分割，然后并列显示。
                    "b_remove_empty_lines": 是否需要将空行去除掉。
    """
    SIMPLE_DICT = "simple_dict"
    COMPLETE_DICT = "complete_dict"
    MATRIX = "matrix"



import pytest
import random
import time
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.computer_science.algorithm.cache_manager.strategy import Strategy_Base
from kevin_toolbox.computer_science.algorithm.cache_manager.variable import CACHE_STRATEGY_REGISTRY


def test_lru_strategy():
    print("test LRU_Strategy")

    strategy = CACHE_STRATEGY_REGISTRY.get(name=":by_last_time:LRU")()  # type: Strategy_Base

    # 写入 1、2、3、4
    #   按照 last_time 排序，优先级为 4、3、2、1
    for i in range(1, 5):
        strategy.notified_by_write_of_cache(key=i, value=i, metadata=None)
    #
    expected_orders = [4, 3, 2, 1]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))
    # 移除并重写 3
    #   优先级变为 3、4、2、1
    strategy.notified_by_remove_of_cache(key=3, metadata=None)
    strategy.notified_by_write_of_cache(key=3, value=3, metadata=None)
    #
    expected_orders = [3, 4, 2, 1]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))
    # 读取 1
    #   优先级变为 1、3、4、2
    strategy.notified_by_read_of_cache(key=1, value=1, metadata=None)
    #
    expected_orders = [1, 3, 4, 2]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))


def test_fifo_strategy():
    print("test FIFO_Strategy")

    strategy = CACHE_STRATEGY_REGISTRY.get(name=":by_initial_time:FIFO")()  # type: Strategy_Base

    # 写入 1、2、3、4
    #   按照 initial_time 排序，优先级为 4、3、2、1
    for i in range(1, 5):
        strategy.notified_by_write_of_cache(key=i, value=i, metadata=None)
    #
    expected_orders = [4, 3, 2, 1]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))

    # 无论后面怎么读取，优先级都不变
    for i in range(1, 5):
        for _ in range(random.randint(3, 6)):
            strategy.notified_by_read_of_cache(key=i, value=i, metadata=None)
    #
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))

    # 重写 1，删除 4
    #   优先级变为 1、3、2
    strategy.notified_by_remove_of_cache(key=1, metadata=None)
    strategy.notified_by_write_of_cache(key=1, value=3, metadata=None)
    strategy.notified_by_remove_of_cache(key=4, metadata=None)
    #
    expected_orders = [1, 3, 2]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))


def test_lfu_strategy():
    print("test LFU_Strategy")

    strategy = CACHE_STRATEGY_REGISTRY.get(name=":by_counts:LFU")()  # type: Strategy_Base

    # 写入 1、2、3、4
    #   按照 counts 排序，优先级为 4==3==2==1
    for i in range(1, 5):
        strategy.notified_by_write_of_cache(key=i, value=i, metadata=None)
    # 读取2次 1，读取1次 3，读取3次 2
    #   counts变为 4:0 3:1 2:3 1:2
    #   优先级变为 2、1、3、4
    for key, counts in {1: 2, 3: 1, 2: 3}.items():
        for i in range(counts):
            strategy.notified_by_read_of_cache(key=key, value=key, metadata=None)
    #
    expected_orders = [2, 1, 3, 4]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))
    # 移除并重写 3，读取一次 4
    #   优先级变为 2、1、4、3
    strategy.notified_by_remove_of_cache(key=3, metadata=None)
    strategy.notified_by_write_of_cache(key=3, value=3, metadata=None)
    strategy.notified_by_read_of_cache(key=4, value=4, metadata=None)
    #
    expected_orders = [2, 1, 4, 3]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))


def test_lst_strategy():
    print("test LST_Strategy")

    strategy = CACHE_STRATEGY_REGISTRY.get(name=":by_survival_time:LST")()  # type: Strategy_Base

    # 写入 1、2、3、4
    #   按照 survival_time 排序，优先级为 4==3==2==1
    for i in range(1, 5):
        strategy.notified_by_write_of_cache(key=i, value=i, metadata={"survival_time": 0})
    init_time = time.time()
    # 依次 读取2次 1，读取1次 3，读取3次 2
    #   优先级变为 2、3、1、4
    for key, counts in {1: 2, 3: 1, 2: 3}.items():
        for i in range(counts):
            strategy.notified_by_read_of_cache(key=key, value=key, metadata={"survival_time": time.time() - init_time})
    #
    expected_orders = [2, 3, 1, 4]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))
    # 移除并重写 3，读取一次 4
    #   优先级变为 4、2、1、3
    strategy.notified_by_remove_of_cache(key=3, metadata=None)
    strategy.notified_by_write_of_cache(key=3, value=3, metadata={"survival_time": 0})
    strategy.notified_by_read_of_cache(key=4, value=4, metadata={"survival_time": time.time() - init_time})
    #
    expected_orders = [4, 2, 1, 3]
    for i in range(0, 5):
        check_consistency(strategy.suggest(refactor_size=i), list(reversed(expected_orders[i:])))



from kevin_toolbox.computer_science.data_structure import Executor
from kevin_toolbox.computer_science.algorithm.pareto_front import get_pareto_points_idx, Direction
import kevin_toolbox.nested_dict_list as ndl
import numpy as np


class Optimum_Picker:
    """
        记录并更新帕累托最优值
            同时支持监控以下行为：
                - 新加值是一个新的帕累托最优值
                - 抛弃一个不再是最优的旧的最优值
            并触发设定的执行器，详见参数 trigger_for_new 和 trigger_for_out
    """

    def __init__(self, **kwargs):
        """
            参数：
                directions:             <list of Direction> 比较的方向
                trigger_for_new:        <Executor> 触发器
                                            当 add() 新添加的监控值是一个新的帕累托最优时，执行该触发器
                                            在执行前，将自动往触发器的 kwargs 中添加 {"metrics": <metrics>, "step": <step>, ...} 等信息
                trigger_for_out:        <Executor> 触发器
                                            当 add() 时候需要抛弃一些不再是的帕累托最优的历史值时，执行该触发器
                                            在执行前，将自动往触发器的 kwargs 中添加 {"metrics": <metrics>, "step": <step>, ...} 等信息
                warmup_steps:           <int> 在经过多少次 add() 之后，再开始比较监控值
                                            默认为 0
                pick_per_steps:         <int> 每经过多少次 add()，就比较一次监控值
                                            默认为 1
        """
        # 默认参数
        paras = {
            "directions": None,
            "trigger_for_new": None,
            "trigger_for_out": None,
            "warmup_steps": 0,
            "pick_per_steps": 1,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert paras["warmup_steps"] >= 0 and paras["pick_per_steps"] >= 1
        paras["directions"] = [Direction(i) for i in paras["directions"]]
        for k in ["trigger_for_out", "trigger_for_new"]:
            assert isinstance(paras[k], (type(None), Executor))

        self.paras = paras
        self._state = self._init_state()

    def _init_state(self):
        return dict(
            optimal_ls=list(),  # [{"metrics":metrics, "record":record, "step":step}, ...]
            step=0,
            b_empty_cache=True,
            last_optimal_nums=0
        )

    def add(self, metrics, b_force_clear_cache=False, **kwargs):
        """
            添加指标

            参数：
                metrics:                    指标
                b_force_clear_cache:        <boolean> 是否强制清空缓存
                                                默认为 False，此时将根据设定的 warmup_steps 和 pick_per_steps 来决定何时清空一次缓存
                **kwargs:                   用户自定义记录
                                                将被添加到 record 中
        """
        assert metrics is not None
        metrics = np.asarray(metrics).reshape(1, -1)
        assert metrics.shape[-1] == len(self.paras["directions"])

        optimal_ls, step = self._state["optimal_ls"], self._state["step"]
        new_record = dict(metrics=metrics, step=step)
        new_record.update(kwargs)
        #
        optimal_ls.append(new_record)
        self._state["step"] += 1

        # warmup & cache
        if not b_force_clear_cache and (step < self.paras["warmup_steps"] or
                                        (step - self.paras["warmup_steps"]) % self.paras["pick_per_steps"] != 0):
            self._state["b_empty_cache"] = False
            return

        # 找出新的帕累托最优值
        points = np.concatenate([i["metrics"] for i in optimal_ls])
        idx_ls = get_pareto_points_idx(points=points, directions=self.paras["directions"])
        idx_ls.sort()
        # 进行触发操作
        if self.paras["trigger_for_new"] is not None:
            for i in filter(lambda i: i >= self._state["last_optimal_nums"], idx_ls):
                self.paras["trigger_for_new"].run(**optimal_ls[i])
        #
        if self.paras["trigger_for_out"] is not None:
            for i in set(range(self._state["last_optimal_nums"])).difference(set(idx_ls)):
                self.paras["trigger_for_out"].run(**optimal_ls[i])

        # 更新
        self._state["optimal_ls"] = [optimal_ls[i] for i in idx_ls]
        self._state["b_empty_cache"] = True
        self._state["last_optimal_nums"] = len(idx_ls)

    def get(self, b_force_clear_cache=False):
        """
            获取当前最优值记录

            参数：
                b_force_clear_cache:        <boolean> 是否强制清空缓存

            返回：
                record_ls:          <list of dict> 最优记录
                b_empty_cache:      <boolean> 缓存是否清空，亦即 record_ls 是否是真正的最优记录
                                        当 pick_per_steps > 1 时，将有部分记录留存在缓存中，没有进行比较，此时的最优记录并不是完整的也不是最新的
        """
        if b_force_clear_cache and not self._state["b_empty_cache"] and len(self._state["optimal_ls"]) > 0:
            # 需要清空缓存，就把最后一次缓存的记录拿出来，重新使用 b_force_clear_cache=False 去 add 一次
            record = self._state["optimal_ls"].pop(-1)
            metrics, step = record.pop("metrics"), record.pop("step")
            self._state["step"] -= 1
            assert self._state["step"] == step
            self.add(metrics=metrics, b_force_clear_cache=True, **record)

        return self._state["optimal_ls"][:], self._state["b_empty_cache"] or b_force_clear_cache

    def clear(self):
        self._state = self._init_state()

    def __len__(self):
        return self._state["step"]

    # ---------------------- 用于保存和加载状态 ---------------------- #
    def load_state_dict(self, state_dict):
        """
            加载状态
        """
        self.clear()
        self._state.update(state_dict)

    def state_dict(self):
        """
            获取状态
        """
        return ndl.copy_(var=self._state, b_deepcopy=True, b_keep_internal_references=True)


if __name__ == '__main__':
    """
        模拟场景
            在训练模型时，要求比较 val_acc_1（maximize） 和 val_error_2（minimize），
            要求保存其帕累托最优时的模型。
    """
    import torch
    import matplotlib.pyplot as plt

    # 一个打乱的圆的采样点序列
    metrics = torch.tensor([(-4.045084971874739, -2.9389262614623632),
                            (-3.1871199487434474, -3.852566213878947),
                            (-2.1288964578253635, 4.524135262330097),
                            (-4.648882429441257, -1.8406227634233896),
                            (-4.648882429441256, 1.8406227634233907),
                            (-0.936906572928623, 4.911436253643443),
                            (0.31395259764656414, -4.990133642141358),
                            (-4.960573506572389, 0.6266661678215226),
                            (-3.1871199487434487, 3.852566213878946),
                            (4.381533400219316, -2.4087683705085805),
                            (0.31395259764656763, 4.990133642141358),
                            (2.6791339748949827, 4.221639627510076),
                            (4.8429158056431545, -1.2434494358242767),
                            (1.5450849718747361, -4.755282581475768),
                            (4.842915805643155, 1.243449435824274),
                            (3.644843137107056, -3.422735529643445),
                            (5.0, 0.0),
                            (-2.128896457825361, -4.524135262330099),
                            (2.6791339748949836, -4.221639627510075),
                            (-4.9605735065723895, -0.6266661678215214),
                            (3.644843137107058, 3.422735529643443),
                            (4.381533400219318, 2.4087683705085765),
                            (-0.9369065729286231, -4.911436253643443),
                            (-4.045084971874736, 2.9389262614623664),
                            (1.5450849718747373, 4.755282581475767)])
    # 右下角的点是帕累托最优
    best_idx_ls = [6, 9, 12, 13, 15, 16, 18]

    # 将x和y坐标分别存储在两个列表中
    x_coords = metrics[:, 0].numpy().tolist()
    y_coords = metrics[:, 1].numpy().tolist()
    # 按顺序绘制点
    plt.plot(x_coords, y_coords, marker='o')
    # 添加顺序标签
    for i, txt in enumerate(range(len(metrics))):
        plt.annotate(txt, (x_coords[i], y_coords[i]), textcoords="offset points", xytext=(0, 5), ha='center')
    plt.show()
    import os
    from kevin_toolbox.data_flow.file import json_
    from kevin_toolbox.patches.for_os import remove

    temp_dir = os.path.join(os.path.dirname(__file__), "temp")
    remove(temp_dir, ignore_errors=True)

    opt_picker = Optimum_Picker(
        warmup_steps=9, pick_per_steps=5,
        trigger_for_new=Executor(
            func=lambda metrics, step: json_.write(metrics.tolist(), os.path.join(temp_dir, f'{step}.json'))),
        trigger_for_out=Executor(func=lambda step, **kwargs: remove(os.path.join(temp_dir, f'{step}.json'))),
        directions=["maximize", "minimize"]
    )
    for s, v in enumerate(metrics):
        opt_picker.add(metrics=v)
        print()
        print(s, v)
        print(opt_picker.get()[1])
        print([i["step"] for i in opt_picker.get()[0]])

    for i in best_idx_ls:
        assert os.path.isfile(os.path.join(temp_dir, f'{i}.json'))



import streamlit as st
from kevin_toolbox.data_flow.file.markdown.table import find_tables
from kevin_toolbox.data_flow.file.markdown.link import find_links
from kevin_toolbox.computer_science.algorithm.for_dict import deep_update
from kevin_toolbox.patches.for_streamlit.markdown import show_image

DEFAULT_DISPLAY_MODE_S = {
    "table_with_image": "by_columns",  # 对于带有图片的表格选择哪种方式显示
    "default": "by_markdown"  # 对于其他表格选择哪种方式显示
}


def _show_table_by_columns(matrix, doc_dir, table_name, **kwargs):
    tab, _ = st.tabs([table_name, "[click to hide table]"])
    with tab:
        for row in matrix:
            col_ls = st.columns(len(row))
            for col, i in zip(col_ls, row):
                with col:
                    show_image(text=i, doc_dir=doc_dir)


METHOD_S = {
    "by_columns": _show_table_by_columns,
    "by_markdown": lambda text, **kwargs: st.markdown(text)
}


def show_table(text, doc_dir=None, display_mode_s=None):
    """
        对 st.markdown 中表格显示部分的改进，具有以下优点
            - 支持显示带有本地图片的表格
            - 支持以下几种方式来显示表格：
                - 用 st.columns 分列分行显示
                - 用 st.markdown 显示（不支持本地图片）
                - 用 st.data_editor 显示（TODO）
    """
    global DEFAULT_DISPLAY_MODE_S, METHOD_S
    display_mode_s = deep_update(stem=DEFAULT_DISPLAY_MODE_S.copy(), patch=display_mode_s if display_mode_s else dict())
    for v in display_mode_s.values():
        assert v in ["by_columns", "by_markdown"]  # "by_data_editor"

    table_ls, part_slices_ls, table_idx_ls = find_tables(text=text, b_compact_format=False)
    for idx, part_slices in enumerate(part_slices_ls):
        part = text[slice(*part_slices)]
        if idx in table_idx_ls:
            table_s = table_ls.pop(0)
            if len(find_links(text=part, b_compact_format=True, type_ls=["image"])) > 0:
                # 带有图片的表格
                method = METHOD_S[display_mode_s["table_with_image"]]
            else:
                method = METHOD_S[display_mode_s["default"]]
            method(text=part, matrix=table_s["matrix"], doc_dir=doc_dir, table_name=f'Table {idx}')
        else:
            # 是表格，且内部无图片，则直接显示
            show_image(text=part, doc_dir=doc_dir)

# 另一种显示表格的方式是通过 data_editor 来显示，但是对图片的显示效果不好
# TODO 可以选择是通过 data_editor 还是 columns，或者原始格式（对本地图片不处理或者使用 base64 代替）来显示表格
# # 创建一个 DataFrame
# data = {
#     'Description': ['This is an image.', "2"],
#     'Image': [f'data:image/png;base64,{convert_image_to_base64(temp)}', temp]  # 使用 Markdown 格式的图片
# }
#
# column_configuration = {
#     "Image": st.column_config.ImageColumn("Avatar", help="The user's avatar", width="large")
# }
#
# import pandas as pd
#
# df = pd.DataFrame(data)
#
# # 创建表格
# # st.table(df)
# st.data_editor(
#     df,
#     column_config=column_configuration,
#     use_container_width=True,
#     hide_index=True,
#     num_rows="fixed"
# )



from .get_rng import get_rng
from .truncated_normal import truncated_normal
from .truncated_multivariate_normal import truncated_multivariate_normal
from .set_rng_state import set_rng_state
from .get_rng_state import get_rng_state



import numpy as np


class DEFAULT_RNG:
    pass


for name in np.random.__all__:
    setattr(DEFAULT_RNG, name, getattr(np.random, name))

func_map_s = {
    'rand': lambda rng, *arg, **kwargs: rng.random(size=arg, **kwargs),
    'randint': 'integers',
    'randn': lambda rng, *arg, **kwargs: rng.normal(size=arg, **kwargs),
    'random_integers': 'integers',
    'random_sample': 'random',
    'ranf': 'random'
}


class My_RNG:
    def __init__(self, rng):
        self._rng = rng

        for name in np.random.__all__:
            setattr(DEFAULT_RNG, name, getattr(np.random, name))

    # self.key
    def __getattr__(self, key):
        if "_rng" not in self.__dict__:
            # _rng 未被设置，未完成初始化。
            return super().__getattr__(key)
        else:
            res = getattr(self._rng, key, None)
            if res is None and key in func_map_s:
                if callable(func_map_s[key]):
                    res = lambda *arg, **kwargs: func_map_s[key](self._rng, *arg, **kwargs)
                else:
                    res = getattr(self._rng, func_map_s[key], None)
            #
            if res is None:
                raise AttributeError(f"attribute '{key}' not found in {type(self)}")
            else:
                return res


def get_rng(seed=None, rng=None, **kwargs):
    if seed is not None:
        # 注意，随机生成器相较于 numpy.random 有部分属性缺失：
        #   ['get_state', 'rand', 'randint', 'randn', 'random_integers', 'random_sample', 'ranf', 'sample', 'seed',
        #   'set_state', 'Generator', 'RandomState', 'SeedSequence', 'MT19937', 'Philox', 'PCG64', 'PCG64DXSM',
        #   'SFC64', 'default_rng', 'BitGenerator']
        rng = My_RNG(rng=np.random.default_rng(seed=seed))
    if rng is not None:
        return rng
    else:
        return DEFAULT_RNG


if __name__ == '__main__':
    a = get_rng(seed=2)

    # 尝试访问随机生成器中缺失的部分方法
    print(a.randn(2, 3))



import os
import warnings
import torch
import cv2
import numpy as np
from PIL import Image
from collections import defaultdict
from kevin_toolbox.data_flow.file import markdown
from kevin_toolbox.patches.for_os.path import replace_illegal_chars, find_illegal_chars
import kevin_toolbox.nested_dict_list as ndl


def save_images_in_ndl(var, plot_dir, doc_dir=None, setting_s=None):
    """
        将ndl结构叶节点下的图片对象保存到 plot_dir 中，并替换为该图片的markdown链接

        参数：
            var:                <dict> 待处理的 ndl 结构
            plot_dir:           <path> 图片保存的目录
            doc_dir:            <path> 输出的markdown文档保存的目录
                                    当有指定时，图片链接将以相对于 doc_dir 的相对路径的形式保存
                                    默认为 None，此时保存的markdown图片链接使用的是绝对路径
            setting_s:          <dict> 配置
                                    指定要在哪些节点下去寻找图片对象，以及转换图片对象时使用的参数
                                    形式为 {<node name>: {"b_is_rgb":<boolean>, ...}, ...}
                                    其中配置项支持：
                                        - b_is_rgb:             待保存的图片是RGB顺序还是BGR顺序
                                        - saved_image_format:   保存图片时使用的格式
                                    默认为 None，此时等效于 {"": {"b_is_rgb": False, "saved_image_format": ".jpg"}}
    """
    if len(find_illegal_chars(file_name=plot_dir, b_is_path=True)) > 0:
        warnings.warn(f'plot_dir {plot_dir} contains illegal symbols, '
                      f'which may cause compatibility issues on certain systems.', UserWarning)
    setting_s = setting_s or {"": {"b_is_rgb": False, "saved_image_format": ".jpg"}}

    # 将配置解释到各个叶节点
    #   从最浅的路径开始，若更深的路径有另外的设置，则以更新的为准
    root_ls = list(setting_s.keys())
    root_ls.sort(key=lambda x: len(ndl.name_handler.parse_name(name=x)[-1]))
    root_to_leaf_s = defaultdict(set)
    leaf_to_root_s = dict()
    leaf_to_value_s = dict()
    for root in root_ls:
        for leaf, v in ndl.get_nodes(var=ndl.get_value(var=var, name=root, b_pop=False), level=-1, b_strict=True):
            leaf = root + leaf
            if leaf in leaf_to_root_s:
                root_to_leaf_s[leaf_to_root_s[leaf]].remove(leaf)
            root_to_leaf_s[root].add(leaf)
            leaf_to_root_s[leaf] = root
            leaf_to_value_s[leaf] = v

    for root, leaf_ls in root_to_leaf_s.items():
        setting_ = setting_s[root]
        for leaf in leaf_ls:
            v = leaf_to_value_s[leaf]
            if isinstance(v, Image.Image):
                v = np.asarray(v)
            elif torch.is_tensor(v):
                v = v.detach().cpu().numpy()
            #
            if isinstance(v, np.ndarray):
                image_path = os.path.join(
                    plot_dir, replace_illegal_chars(
                        file_name=f'{leaf}_{setting_["saved_image_format"]}', b_is_path=False)
                )
                os.makedirs(os.path.dirname(image_path), exist_ok=True)
                if setting_["b_is_rgb"]:
                    v = cv2.cvtColor(v, cv2.COLOR_RGB2BGR)
                cv2.imwrite(image_path, v)
                v_new = markdown.generate_link(
                    name=os.path.basename(image_path),
                    target=os.path.relpath(image_path, doc_dir) if doc_dir is not None else image_path, type_="image")
            elif v is None:
                v_new = "/"
            else:
                v_new = v
            ndl.set_value(var=var, name=leaf, b_force=False, value=v_new)
    return var






import numpy as np
import torch
import warnings
import kevin_toolbox.nested_dict_list as ndl


def check_consistency(*args, tolerance=1e-7, require_same_shape=True):
    """
        检查 args 中多个变量之间是否一致
            变量支持 python 的所有内置类型，以及复杂的 nested_dict_list 结构， array 等
            对于 array，不区分 numpy 的 array，torch 的 tensor，还是 tuple of number，只要其中的值相等，即视为相同。

        参数：
            tolerance:          <float> 判断 <np.number/np.bool_> 之间是否一致时，的容许误差。
                                    默认为 1e-7。
            require_same_shape: <boolean> 是否强制要求 array 变量的形状一致。
                                    默认为 True，
                                    当设置为 False 时，不同形状的变量可能因为 numpy 的 broadcast 机制而在比较前自动 reshape 为相同维度，进而可能通过比较。
    """
    assert len(args) >= 2

    # 复杂结构 ndl
    if isinstance(args[0], (list, dict,)):
        nodes_ls = [sorted(ndl.get_nodes(var=arg, level=-1), key=lambda x: x[0]) for arg in args]
        names_ls, values_ls = [], []
        for nodes in nodes_ls:
            names_ls.append([i[0] for i in nodes])
            values_ls.append([i[1] for i in nodes])
        try:
            _check_item(*names_ls, tolerance=tolerance, require_same_shape=True)
        except AssertionError as e:
            raise AssertionError(f'inputs <nested_dict_list> has different structure\nthe nodes that differ are:\n{e}')
        for its in zip(names_ls[0], *values_ls):
            try:
                _check_item(*its[1:], tolerance=tolerance, require_same_shape=require_same_shape)
            except AssertionError as e:
                raise AssertionError(
                    f'value of nodes {its[0]} in inputs <nested_dict_list> are inconsistent\nthe difference is:\n{e}')
    # 简单结构
    else:
        _check_item(*args, tolerance=tolerance, require_same_shape=require_same_shape)


def _check_item(*args, tolerance, require_same_shape):
    """
        检查 args 中多个 array 之间是否一致

        工作流程：
            1. 对于 args 都是 tuple/list 且每个 tuple/list 的长度都一致的情况，将会拆分为对应各个元素递归进行比较。
            2. 将 args 中的 tuple 和 tensor 分别转换为 list 和 numpy。
            3. 检查 args 中是否有 np.array 或者 tensor，若有则根据 require_same_shape 判断其形状是否一致。
            4. 先将输入的 args 中的所有变量转换为 np.array;
                然后使用 issubclass() 判断转换后得到的变量属于以下哪几种基本类型：
                    - 当所有变量都属于 np.number 数值（包含int、float等）或者 np.bool_ 布尔值时，
                        将对变量两两求差，当差值小于给定的容许误差 tolerance 时，视为一致。
                        注意：在比较过程中，若变量中存在 np.nan 值，将会首先比较有 np.nan 值的位置是否相等，然后再比较非 np.nan 值部分。
                            亦即在相同位置上都具有 np.nan 视为相同。
                            比如 [np.nan, np.nan, 3] 和 [np.nan, np.nan, 3] 会视为相等，
                            [np.nan, np.nan] 和 np.nan 在 require_same_shape=False 时会被视为相等。
                    - 当所有变量都属于 np.flexible 可变长度类型（包含string等）或者 np.object 时，
                        将使用==进行比较，当返回值都为 True 时，视为一致。
                    - 当变量的基本类型不一致（比如同时有np.number和np.flexible）时，
                        直接判断为不一致。
                numpy 中基本类型之间的继承关系参见： https://numpy.org.cn/reference/arrays/scalars.html

        参数：
            tolerance:          <float> 判断 <np.number/np.bool_> 之间是否一致时，的容许误差。
            require_same_shape: <boolean> 是否强制要求变量的形状一致。
                                    注意：仅在原始 args 中含有 np.array 或者 tensor 的情况会采取 broadcast，亦即此时该参数才会起效。
                                    当设置为 False 时，不同形状的变量可能因为 broadcast 机制而在比较前自动 reshape 为相同维度，进而通过比较。
    """
    warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
    warnings.filterwarnings("ignore", category=DeprecationWarning)

    assert len(args) >= 2
    assert isinstance(tolerance, (int, float,)) and tolerance >= 0
    raw_args = args

    # 当 args 都是 tuple/list 且每个 tuple 的长度都一致时，将会拆分为对应各个元素进行比较
    if all([isinstance(i, (tuple, list)) for i in args]) and all([len(i) == len(args[0]) for i in args[1:]]):
        for i, it_ls in enumerate(zip(*args)):
            try:
                _check_item(*it_ls, tolerance=tolerance, require_same_shape=require_same_shape)
            except AssertionError as e:
                raise AssertionError(f'elements {i} are inconsistent\nthe difference is:\n{e}')
        return

    #
    args = ndl.traverse(var=list(args), match_cond=lambda _, __, v: isinstance(v, (tuple,)), action_mode="replace",
                        converter=lambda _, v: list(v), b_traverse_matched_element=True)
    args = ndl.traverse(var=list(args), match_cond=lambda _, __, v: torch.is_tensor(v), action_mode="replace",
                        converter=lambda _, v: v.detach().cpu().numpy(), b_traverse_matched_element=False)
    b_has_raw_array = any([isinstance(i, np.ndarray) for i in args])

    try:
        args = [np.asarray(v) for v in args]  # if b_has_raw_array else [np.array(v, dtype=object) for v in args]
    except Exception as e:
        raise RuntimeError(f'{raw_args} cannot be converted to np.array, \n'
                           f'because {e}')

    # 比较形状
    if b_has_raw_array:
        if require_same_shape:
            # 要求形状一致
            for v in args[1:]:
                assert args[0].shape == v.shape, \
                    f"{args[0]}, {v}, different shape: {args[0].shape}, {v.shape}"
        else:
            # 否则要求至少能够进行 broadcast
            for v in args[1:]:
                try:
                    np.broadcast_arrays(args[0], v)
                except:
                    raise AssertionError(f'{args[0]}, {v}, failed to broadcast')
            # 如果都是空的 array，直接视为相等
            if all([i.size == 0 for i in args]):
                return
    b_allow_broadcast = b_has_raw_array and not require_same_shape

    # 比较值
    if issubclass(args[0].dtype.type, (np.number, np.bool_,)):
        # 数字类型
        for v in args[1:]:
            assert issubclass(v.dtype.type, (np.number, np.bool_,))
            v_0, v_1 = args[0].astype(dtype=float), v.astype(dtype=float)
            v_0, v_1 = np.broadcast_arrays(v_0, v_1) if b_allow_broadcast else (v_0, v_1)
            assert v_0.shape == v_1.shape, \
                f'{v_0}, {v_1}, different shape: {v_0.shape}, {v_1.shape}'
            #
            if v_0.size > 0:
                try:
                    if np.any(np.isnan(v)):
                        assert np.all(np.isnan(v_0) == np.isnan(v_1))
                        v_0 = np.nan_to_num(v_0, nan=1e10)
                        v_1 = np.nan_to_num(v_1, nan=1e10)
                    assert np.max(np.abs(v_0 - v_1)) < tolerance
                except AssertionError:
                    raise AssertionError(f"{args[0]}, {v}, deviation: {np.max(np.abs(args[0] - v))}")
    elif issubclass(args[0].dtype.type, (np.flexible, object,)):
        # 可变长度类型
        for v in args[1:]:
            assert issubclass(v.dtype.type, (np.flexible, object,))
            v_0, v_1 = np.broadcast_arrays(args[0], v) if b_allow_broadcast else (args[0], v)
            assert v_0.shape == v_1.shape, \
                f'{v_0}, {v_1}, different shape: {v_0.shape}, {v_1.shape}\n' + (
                    '' if require_same_shape else
                    f'\tMore details: \n'
                    f'\t\tAlthough require_same_shape=False has been setted, broadcast failed because the variable at \n'
                    f'\t\tthis position does not contain elements of type np.array and tensor.')
            #
            for i, j in zip(v_0.reshape(-1), v_1.reshape(-1)):
                if i is j:
                    continue
                temp = i == j
                if isinstance(temp, (bool,)):
                    assert temp, \
                        f"{args[0]}, {v}, diff: {temp}"
                else:
                    assert temp.all(), \
                        f"{args[0]}, {v}, diff: {temp}"
    else:
        raise ValueError


if __name__ == '__main__':
    a = np.array([[1, 2, 3]])
    b = np.array([[1, 2, 3]])
    c = {'d': 3, 'c': 4}

    # var = ((1, 2), (4))
    #
    # var = ndl.traverse(var=[var], match_cond=lambda _, __, v: isinstance(v, (tuple,)), action_mode="replace",
    #                    converter=lambda _, v: list(v), b_traverse_matched_element=True)[0]
    # print(var)



from collections import defaultdict
from kevin_toolbox.nested_dict_list import get_nodes, get_value, set_value
from kevin_toolbox.nested_dict_list import value_parser


def replace_identical_with_reference(var, flag="same", match_cond=None, b_reverse=False):
    """
        将具有相同 id 的多个节点，替换为单个节点和其多个引用的形式
            一般用于去除冗余部分，压缩 ndl 的结构

        参数：
            var:
            flag:                   <str> 引用标记头，表示该节点应该替换为指定节点的内容
                                        默认为 "same"
                                        注意，在正向过程中，如果遇到本身就是以该 flag 开头的字符串，会自动在前添加多一个 flag 以示区分，
                                            然后在逆向过程中，遇到两个 flag 标记开头的字符串将删除一个，然后跳过不处理。
            match_cond:             <func> 仅对匹配上（返回True视为匹配上）的节点进行处理
                                        函数类型为 def(name, value)
                                        其中：
                                            name            该节点在结构体中的位置
                                            value           节点的值
                                        默认不对 int、float、bool、str、None 等类型进行处理
            b_reverse:              <boolean> 是否进行逆向操作
    """
    if match_cond is None:
        match_cond = lambda name, value: not isinstance(value, (int, float, bool, str, type(None)))
    assert callable(match_cond)

    if b_reverse:
        return _reverse(var, flag)
    else:
        return _forward(var, flag, match_cond)


def _forward(var, flag, match_cond):
    id_to_height_s = defaultdict(set)
    id_to_name_s = defaultdict(set)
    height = 1
    while True:
        node_ls = get_nodes(var=var, level=-height, b_strict=True)
        if not node_ls:
            break
        for name, value in node_ls:
            if not match_cond(name, value):
                continue
            id_to_name_s[id(value)].add(name)
            id_to_height_s[id(value)].add(height)
        height += 1

    #
    for k, v in list(id_to_height_s.items()):
        # 仅保留有多个节点对应的 id
        if len(id_to_name_s[k]) <= 1:
            id_to_height_s.pop(k)
            id_to_name_s.pop(k)
            continue
        # 具有相同 id 的节点所处的高度应该相同
        assert len(v) == 1, f'nodes {id_to_name_s[k]} have different heights: {v}'
    # 按高度排序
    id_vs_height = sorted([(k, v.pop()) for k, v in id_to_height_s.items()], key=lambda x: x[1], reverse=True)

    # 从高到低，依次将具有相同 id 的节点替换为 单个节点和多个引用 的形式
    temp = []
    processed_name_set = set()
    for k, _ in id_vs_height:
        # 找出父节点仍然未被处理的节点（亦即仍然能够访问到的节点）
        unprocessed_name_set = {n for n in id_to_name_s[k] if id(get_value(var=var, name=n, default=temp)) == k}
        if len(unprocessed_name_set) <= 1:
            continue
        # 任选其一进行保留，其余改为引用
        keep_name = unprocessed_name_set.pop()
        for name in unprocessed_name_set:
            try:
                var = set_value(var=var, name=name, value=f'<{flag}>{{{keep_name}}}', b_force=False)
            except:
                breakpoint()
        processed_name_set.update(unprocessed_name_set)

    # 将叶节点中，未被处理过，且是 str，且以 flag 开头的字符串，添加多一个 flag，以示区分
    for name, value in get_nodes(var=var, level=-1, b_strict=True):
        if name not in processed_name_set and isinstance(value, str) and value.startswith(f'<{flag}>'):
            var = set_value(var=var, name=name, value=f'<{flag}>' + value, b_force=False)

    return var


class _My_Str:
    def __init__(self, s):
        self.s = s


def _reverse(var, flag):
    # 找出叶节点中，带有2个以上 flag 标记的字符串，删除其中一个标记，并使用 _My_Str 包裹，以便与普通引用节点区分
    for name, value in get_nodes(var=var, level=-1, b_strict=True):
        if isinstance(value, str) and value.startswith(f'<{flag}><{flag}>'):
            var = set_value(var=var, name=name, value=_My_Str(value[len(flag) + 2:]), b_force=False)
    # 解释引用
    var, _ = value_parser.parse_and_eval_references(var=var, flag=flag)
    # 解除 _My_Str 包裹
    for name, value in get_nodes(var=var, level=-1, b_strict=True):
        if isinstance(value, _My_Str):
            var = set_value(var=var, name=name, value=value.s, b_force=False)
    return var


if __name__ == '__main__':
    import numpy as np
    from kevin_toolbox.nested_dict_list import copy_

    a = np.array([1, 2, 3])
    b = np.ones((2, 3))
    c = [a, b]
    d = {"a": a, "b": b}
    e = {"c1": c, "c2": c}
    x = [e, a, d, c, "<same>{@1}", "<same><same>{@1}"]

    print(x)

    y = replace_identical_with_reference(var=copy_(x, b_deepcopy=True), flag="same")
    print(y)

    x1 = replace_identical_with_reference(var=y, flag="same", b_reverse=True)
    print(x1)

    from kevin_toolbox.patches.for_test import check_consistency

    check_consistency(x, x1)



import os
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from kevin_toolbox.patches.for_matplotlib.color import generate_color_list
from kevin_toolbox.patches.for_os.path import replace_illegal_chars


def plot_scatters_matrix(data_s, title, x_name_ls, cate_name=None, output_dir=None, cate_color_s=None, **kwargs):
    paras = {
        "dpi": 200,
        "diag_kind": "kde"  # 设置对角线图直方图/密度图 {‘hist’, ‘kde’}
    }
    assert cate_name in data_s and len(set(x_name_ls).difference(set(data_s.keys()))) == 0
    if cate_color_s is None:
        temp = set(data_s[cate_name])
        cate_color_s = {k: v for k, v in zip(temp, generate_color_list(len(temp)))}
    assert set(cate_color_s.keys()) == set(data_s[cate_name])
    paras.update(kwargs)

    plt.clf()
    # 使用seaborn绘制散点图矩阵
    sns.pairplot(
        pd.DataFrame(data_s),
        diag_kind=paras["diag_kind"],  # 设置对角线图直方图/密度图 {‘hist’, ‘kde’}
        hue=cate_name,  # hue 表示根据该列的值进行分类
        palette=cate_color_s, x_vars=x_name_ls, y_vars=x_name_ls,  # x_vars，y_vars 指定子图的排列顺序
    )
    #
    plt.subplots_adjust(top=0.95)
    plt.suptitle(f'{title}', y=0.98, x=0.47)
    # g.fig.suptitle(f'{title}', y=1.05)

    if output_dir is None:
        plt.show()
        return None
    else:
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, f'{replace_illegal_chars(title)}.png')
        plt.savefig(output_path, dpi=paras["dpi"])
        return output_path


if __name__ == '__main__':
    data_s_ = dict(
        x=[1, 2, 3, 4, 5],
        y=[2, 4, 6, 8, 10],
        z=[2, 4, 6, 8, 10],
        categories=['A', 'B', 'A', 'B', 'A'],
        title='test',
    )

    plot_scatters_matrix(data_s=data_s_, title='test', x_name_ls=['y', 'x', 'z'], cate_name='categories',
                         cate_color_s={'A': 'red', 'B': 'blue'})



def escape_tuple_and_set(x):
    """
        将 tuple 和 set 进行转义
            转义：     x ==> f"<eval>{x}"
            反转义：   f"<eval>{x}" ==> x

        为什么要进行转义？
            由于 json 中会将 tuple 作为 list 进行保存，同时也无法保存 set，因此在保存过程中会丢失相应信息。
    """
    if isinstance(x, (tuple, set,)) or (isinstance(x, (str,)) and x.startswith("<eval>")):
        return f'<eval>{x}'
    else:
        return x


if __name__ == '__main__':
    print(escape_tuple_and_set((1, 2, "\'1\'")))
    # <eval>(1, 2, "'1'")
    print(escape_tuple_and_set({"1", (1, 2, 3), 233}))
    # <eval>{'1', 233, (1, 2, 3)}

    print(escape_tuple_and_set("<eval>233"))
    # <eval><eval>233



def matrix_to_complete(matrix, orientation, chunk_size=None, chunk_nums=None, b_remove_empty_lines=False):
    """
        将二维数组形式的 MATRIX 格式（比如find_tables()的返回列表的元素），转换成 COMPLETE_DICT 格式

        参数：
            matrix:                     <list of row> 二维数组形式的表格
            orientation:                <str> 解释表格时取哪个方向
                                            支持以下值：
                                            "vertical" / "v":       将第一行作为标题
                                            "horizontal" / "h":     将第一列作为标题
            chunk_nums:                 <int> 表格被平均分割为多少份进行并列显示。
            chunk_size:                 <int> 表格被按照最大长度进行分割，然后并列显示。
                以上两个参数是用于解释 generate_table() 中使用对应参数生成的表格，其中 chunk_size 仅作检验行数是否符合要求，
                对解释表格无作用。但是当指定该参数时，将视为表格有可能是多个表格并列的情况，因此将尝试根据标题的重复规律，
                推断出对应的 chunk_nums，并最终将其拆分成多个表格。
            b_remove_empty_lines:       <boolean> 移除空的行、列
    """
    # 检验参数
    assert chunk_nums is None or 1 <= chunk_nums
    assert chunk_size is None or 1 <= chunk_size
    assert isinstance(matrix, (list, tuple,))
    assert orientation in ["vertical", "horizontal", "h", "v"]

    # 转换为字典形式
    if orientation not in ["vertical", "v"]:
        # 需要转为垂直方向
        matrix = list(zip(*matrix))
    r_nums, c_nums = len(matrix), len(matrix[0])
    if chunk_size is not None:
        assert chunk_size == r_nums - 1, \
            (f'The number of values {r_nums - 1} actually contained in the table '
             f'does not match the specified chunk_size {chunk_size}')
        chunk_nums = c_nums // _find_shortest_repeating_pattern_size(arr=matrix[0])
    chunk_nums = 1 if chunk_nums is None else chunk_nums
    assert c_nums % chunk_nums == 0, \
        f'The number of headers actually contained in the table does not match the specified chunk_nums, ' \
        f'Expected n*{chunk_nums}, but got {c_nums}'
    # 解释出标题
    keys = matrix[0][0:c_nums // chunk_nums]
    # 解释出值
    if chunk_nums == 1:
        values = matrix[1:]
    else:
        values = []
        for i in range(chunk_nums):
            for j in range(1, r_nums):
                values.append(matrix[j][i * len(keys):(i + 1) * len(keys)])
    # 去除空行
    if b_remove_empty_lines:
        values = [line for line in values if any(i != '' for i in line)]
    table_s = {i: {"title": k, "values": list(v)} for i, (k, v) in enumerate(zip(keys, list(zip(*values))))}
    # 去除空列
    if b_remove_empty_lines:
        table_s = {k: v_s for k, v_s in table_s.items() if v_s["title"] != '' and any(i != '' for i in v_s["values"])}

    return table_s


def _find_shortest_repeating_pattern_size(arr):
    n = len(arr)

    # 部分匹配表
    pi = [0] * n
    k = 0
    for i in range(1, n):
        if k > 0 and arr[k] != arr[i]:
            k = 0
        if arr[k] == arr[i]:
            k += 1
        pi[i] = k

    # 最短重复模式的长度
    pattern_length = n - pi[n - 1]
    # 是否是完整的重复模式
    if n % pattern_length != 0:
        pattern_length = n
    return pattern_length


if __name__ == '__main__':
    from kevin_toolbox.data_flow.file.markdown import find_tables

    # # 示例Markdown表格文本
    # file_path = ""
    # with open(file_path, 'r') as f:
    #     markdown_text = f.read()

    # markdown_text = """
    # | Name | Age | Occupation |
    # |------|-----|------------|
    # | Alice | 28  | Engineer   |
    # | Bob   | 23  | Teacher    |
    # | Name | Age | Occupation |
    # | Carol | 32  | Hacker   |
    # | David | 18  | Student   |
    # """

    markdown_text = """
|  | a | b |  | a | b |  | a | b |
| --- | --- | --- | --- | --- | --- | --- | --- | --- |
|  | 0 | 2 |  | 4 | 6 |  | 7 | 9 |
|  | 1 | 3 |  | 5 | 7 |  | 8 | : |
|  | 2 | 4 |  | 6 | 8 |  | 9 | ; |
|  | 3 | 5 |  |  |  |  |  |  |
"""
    table_ls = find_tables(text=markdown_text)

    # 调用函数并打印结果
    tables = matrix_to_complete(matrix=table_ls[0], orientation="v", chunk_nums=3, b_remove_empty_lines=True)
    print(tables)



import pytest
import time
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.data_flow.core.cache import Cache_Manager_for_Iterator


def test_cache_manager_for_iterator():
    print("test Cache_Manager_for_Iterator")

    cache_manager = Cache_Manager_for_Iterator(
        iterator=range(10),
        b_del_cache_when_exit=True,
        paras_for_memo_cache=dict(upper_bound=3, refactor_size=3, strategy=":by_last_time:LRU")
    )

    check_consistency(
        cache_manager.file_dict,
        {i: f'{i}.pkl' for i in range(10)}
    )
    check_consistency(
        cache_manager.memo_cache_manager.metadata_s,
        dict()
    )
    key_to_counts = {0: 3, 3: 2, 6: 1, 9: 4}
    for key, counts in key_to_counts.items():
        for _ in range(counts):
            time.sleep(0.05)
            check_consistency(cache_manager.read(key), key)
    # counts
    for key, counts in key_to_counts.items():
        if key == 0:
            assert key not in cache_manager.memo_cache_manager.metadata_s
        else:
            check_consistency(cache_manager.memo_cache_manager.metadata_s[key]["counts"], counts - 1)



import heapq
from kevin_toolbox.computer_science.algorithm.cache_manager.strategy import Strategy_Base, LFU_Strategy
from kevin_toolbox.computer_science.algorithm.cache_manager.strategy.lfu_strategy import Score_to_Key
from kevin_toolbox.computer_science.algorithm.cache_manager.variable import CACHE_STRATEGY_REGISTRY


@CACHE_STRATEGY_REGISTRY.register()
class LST_Strategy(LFU_Strategy):
    """
        删除访问频率最低的部分
        drop items with smaller survival_time
    """

    name = ":by_survival_time:LST_Strategy"

    def notified_by_write_of_cache(self, key, value, metadata):
        self.record_s[key] = Score_to_Key(score=metadata["survival_time"], key=key)
        heapq.heappush(self.order_ls, self.record_s[key])

    def notified_by_read_of_cache(self, key, value, metadata):
        self.record_s[key].score = metadata["survival_time"]


# 添加其他别名
for name in [":by_survival_time:LST", ":by_survival_time:Least_Survival_Time", ":by_survival_time:drop_smaller"]:
    CACHE_STRATEGY_REGISTRY.add(obj=LST_Strategy, name=name, b_force=False, b_execute_now=False)



import os
import re

ILLEGAL_CHARS = {
    '<': '＜',
    '>': '＞',
    ':': '：',
    '"': '＂',
    '/': '／',
    '\\': '＼',
    '|': '｜',
    '?': '？',
    '*': '＊'
}
ILLEGAL_CHARS_PATTERN = re.compile('|'.join(re.escape(char) for char in ILLEGAL_CHARS.keys()))


def replace_illegal_chars(file_name, b_is_path=False):
    """
        将给定的文件名/路径中的非法符号替换为合法形式
            所谓非法符号是指在特定系统（win/mac等）中文件名不允许出现的字符，
            不建议在任何系统下使用带有这些符号的文件名，即使这些符号在当前系统中是合法的，
            以免在跨系统时出现兼容性问题。

        参数：
            file_name:      <str>
            b_is_path:      <bool> 是否将file_name视为路径
                                默认为 False
                                当设置为 True 时，将会首先将file_name分割，再逐级处理目录名，最后合并为路径
    """
    if not b_is_path:
        res = _replace_illegal_chars(var=file_name)
    else:
        temp = file_name.split(os.sep, -1)
        res = os.path.join(*[_replace_illegal_chars(var=i) for i in temp if len(i) > 0])
        if len(temp[0]) == 0:
            res = os.sep + res
    return res


def _replace_illegal_chars(var):
    global ILLEGAL_CHARS_PATTERN
    return ILLEGAL_CHARS_PATTERN.sub(lambda m: ILLEGAL_CHARS[m.group(0)], var)


if __name__ == '__main__':
    file_path = 'data0//b/<?>.md'
    print(replace_illegal_chars(file_name=file_path, b_is_path=True))
    print(replace_illegal_chars(file_name=file_path, b_is_path=False))



def clear_border_of_axes(ax):
    """
        用于清除 ax 中的坐标轴和 ticks
    """
    ax.set_xticks([])
    ax.set_yticks([])
    ax.spines['left'].set_color('none')
    ax.spines['right'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['top'].set_color('none')
    return ax


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    #
    fig, ax = plt.subplots()

    x = [1, 4]
    y = [1, 10]
    ax.plot(x, y)

    # 设置坐标轴的范围，以便更好地展示直线
    ax.set_xlim([0, 5])
    ax.set_ylim([0, 15])

    # 添加标题和坐标轴标签
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    clear_border_of_axes(ax)

    # 显示图形
    plt.show()



import os
import streamlit as st
from kevin_toolbox.data_flow.file.markdown.link import find_links


def show_image(text, doc_dir=None):
    """
        对 st.markdown 中图片显示部分的改进，具有以下优点
            - 能够正确显示本地的图片，以 st.image 方式或者 base64 方式（待实现 TODO）
    """
    link_ls, part_slices_ls, link_idx_ls = find_links(text=text, b_compact_format=False, type_ls=["image"])
    for i, part_slices in enumerate(part_slices_ls):
        if i in link_idx_ls:
            link_s = link_ls.pop(0)
            st.image(image=os.path.join(doc_dir, link_s["target"]) if doc_dir else link_s["target"],
                     caption=link_s["name"] or link_s["title"])
        else:
            st.markdown(text[slice(*part_slices)])

# from PIL import Image
# from io import BytesIO
# import base64
#
# def convert_image_to_base64(file_path=None, image=None, output_format="png"):
#     """
#         将图片转为 base64 编码的字符串
#     """
#     assert output_format in ["png", "jpeg"]
#     if file_path:
#         image = Image.open(file_path)
#     assert image is not None
#     with BytesIO() as buffer:
#         image.save(buffer, 'png')  # or 'jpeg'
#         res = base64.b64encode(buffer.getvalue()).decode('utf-8')
#     return res
#
#
# if __name__ == "__main__":
#     image_path = "/home/SENSETIME/xukaiming/Desktop/gitlab_repos/face_liveness_datasets/deploy_for_streamlit/pages/test/test_data/images/7.jpg"
#     print(convert_image_to_base64(image_path))



import numpy as np


def entropy(pdf, b_need_normalize=False):
    """
        参数：
            pdf:                    <list/array> 概率分布
            b_need_normalize:       <boolean> 是否将输出的熵归一化到 0~1 区间
                                        默认为 False，此时输出的熵遵从原始定义，其大小范围与输出概率分布的维度 n 有关，为 0 ~ log_{2}(n)
                                        当设置为 True，将会除以 log_{2}(n) 以进行归一化。
    """
    pdf = np.asarray(pdf, dtype=float).reshape(-1)
    # 将概率分布数组中小于等于0的元素设为1，避免出现log(0)的情况
    pdf = np.maximum(pdf, 1e-15)
    #
    pdf /= np.sum(pdf)
    # 计算熵值
    res = -np.sum(pdf * np.log2(pdf))
    if b_need_normalize:
        res /= np.log2(len(pdf))
    return res


if __name__ == '__main__':
    print(entropy(pdf=[0.1, 0.1, 0.7, 0.1], b_need_normalize=True))  # 0.6783898247235198
    print(entropy(pdf=[0.05, 0.05, 0.05, 0.05, 0.8], b_need_normalize=True))  # 0.4831881303119284



from .accumulator_base import Accumulator_Base
from .exponential_moving_average import Exponential_Moving_Average
from .average_accumulator import Average_Accumulator
from .accumulator_for_ndl import Accumulator_for_Ndl



import os
import copy
from kevin_toolbox.computer_science.algorithm import for_seq
import matplotlib.pyplot as plt
from kevin_toolbox.patches.for_os.path import replace_illegal_chars

# TODO 在 linux 系统下遇到中文时，尝试自动下载中文字体，并尝试自动设置字体
# font_path = os.path.join(root_dir, "utils/SimHei.ttf")
# font_name = FontProperties(fname=font_path)


def plot_bars(data_s, title, x_name, y_label=None, output_dir=None, **kwargs):
    data_s = copy.deepcopy(data_s)
    paras = {
        "dpi": 200
    }
    paras.update(kwargs)

    plt.clf()
    #
    x_all_ls = data_s.pop(x_name)
    #
    for i, (k, y_ls) in enumerate(data_s.items()):
        if i == 0:
            plt.bar([j - 0.1 for j in range(len(x_all_ls))], y_ls, width=0.2, align='center', label=k)
        else:
            plt.bar([j + 0.1 for j in range(len(x_all_ls))], y_ls, width=0.2, align='center', label=k)

    plt.xlabel(f'{x_name}')
    plt.ylabel(f'{y_label if y_label else "value"}')
    temp = for_seq.flatten_list([list(i) for i in data_s.values()])
    y_min, y_max = min(temp), max(temp)
    plt.ylim(max(min(y_min, 0), y_min - (y_max - y_min) * 0.2), y_max + (y_max - y_min) * 0.1)
    plt.xticks(list(range(len(x_all_ls))), labels=x_all_ls)  # , fontproperties=font_name
    plt.title(f'{title}')
    # 显示图例
    plt.legend()

    if output_dir is None:
        plt.show()
        return None
    else:
        os.makedirs(output_dir, exist_ok=True)
        output_path = os.path.join(output_dir, f'{replace_illegal_chars(title)}.png')
        plt.savefig(output_path, dpi=paras["dpi"])
        return output_path


if __name__ == '__main__':
    plot_bars(data_s={
        'a': [1.5, 2, 3, 4, 5],
        'b': [5, 4, 3, 2, 1],
        'c': [1, 2, 3, 4, 5]},
        title='test', x_name='a', output_dir=os.path.join(os.path.dirname(__file__), "temp"))



from kevin_toolbox.patches.for_matplotlib.color import Color_Format, get_format


def hex_to_rgba(hex_color):
    hex_color = hex_color.lstrip('#')
    assert len(hex_color) in (6, 8), \
        f'hex_color should be 6 or 8 characters long (not including #). but got {len(hex_color)}'
    res = list(int(hex_color[i * 2:i * 2 + 2], 16) for i in range(len(hex_color) // 2))
    if len(res) not in (3, 4):
        breakpoint()
    if len(res) == 4:
        res[3] /= 255
    return tuple(res)


def rgba_to_hex(rgba):
    assert len(rgba) in (3, 4), \
        f'rgba should be 3 or 4 elements long. but got {len(rgba)}'
    if len(rgba) == 4:
        rgba = list(rgba)
        rgba[3] = max(0, min(255, int(255 * rgba[3])))
    res = "#"
    for i in rgba:
        res += f'{i:02X}'
    return res


NAME_TO_HEX = {
    'blue': '#0000FF',
    'red': '#FF0000',
    'green': '#008000',
    'orange': '#FFA500',
    'purple': '#800080',
    'yellow': '#FFFF00',
    'brown': '#A52A2A',
    'pink': '#FFC0CB',
    'gray': '#808080',
    'olive': '#808000',
    'cyan': '#00FFFF'
}
HEX_TO_NAME = {v: k for k, v in NAME_TO_HEX.items()}


def natural_name_to_hex(name):
    global NAME_TO_HEX
    name = name.lower()
    assert name in NAME_TO_HEX, \
        f'{name} is not a valid color name.'
    return NAME_TO_HEX[name]


def hex_to_natural_name(hex_color):
    global HEX_TO_NAME
    hex_color = hex_color.upper()[:7]
    assert hex_color in HEX_TO_NAME, \
        f'{hex_color} does not has corresponding color name.'
    return HEX_TO_NAME[hex_color]


CONVERT_PROCESS_S = {
    (Color_Format.HEX_STR, Color_Format.NATURAL_NAME): hex_to_natural_name,  # (from, to): process
    (Color_Format.HEX_STR, Color_Format.RGBA_ARRAY): hex_to_rgba,
    (Color_Format.NATURAL_NAME, Color_Format.HEX_STR): natural_name_to_hex,
    (Color_Format.NATURAL_NAME, Color_Format.RGBA_ARRAY): lambda x: hex_to_rgba(natural_name_to_hex(x)),
    (Color_Format.RGBA_ARRAY, Color_Format.HEX_STR): rgba_to_hex,
    (Color_Format.RGBA_ARRAY, Color_Format.NATURAL_NAME): lambda x: hex_to_natural_name(rgba_to_hex(x))
}


def convert_format(var, output_format, input_format=None):
    """
        在各种颜色格式之间进行转换

        参数：
            var:
            input_format:       <str> 描述输入的格式。
                                    支持 HEX_STR、NATURAL_NAME、RGBA_ARRAY 等格式，
                                    默认为 None，此时将根据输入推断格式
            output_format:      <str/list of str> 输出的目标格式。
                                    当输入是一个 tuple/list 时，将输出其中任一格式，具体规则为：
                                        - 当 input_format 不在可选的输出格式中时，优先按照第一个输出格式进行转换。
                                            若转换失败，则按照第二个输出格式进行转换。依次类推。
                                        - 当 input_format 在可选的输出格式中时，不进行转换。
    """
    global CONVERT_PROCESS_S
    if input_format is None:
        input_format = get_format(var=var)
    input_format = Color_Format(input_format)
    if not isinstance(output_format, (list, tuple,)):
        output_format = [output_format]
    output_format = [Color_Format(i) for i in output_format]

    if input_format in output_format:
        return var
    else:
        for output_format_i in output_format:
            try:
                return CONVERT_PROCESS_S[(input_format, output_format_i)](var)
            except Exception as e:
                raise Exception(f'fail to convert {var} from {input_format} to {output_format}, beacause: {e}')


if __name__ == '__main__':
    print(hex_to_rgba('#FF57337F'))
    print(rgba_to_hex((255, 87, 51, 0.5)))
    print(natural_name_to_hex('pink'))
    print(convert_format(var='#FF57337F', input_format='hex_str', output_format='rgba_array'))
    print(convert_format(var="#0000FF", output_format="rgba_array"))



import heapq
from kevin_toolbox.computer_science.algorithm.cache_manager.strategy import Strategy_Base
from kevin_toolbox.computer_science.algorithm.cache_manager.variable import CACHE_STRATEGY_REGISTRY


class Score_to_Key:
    def __init__(self, score, key):
        self.score = score
        self.key = key

    def __lt__(self, other):
        return self.score < other.score

    def __eq__(self, other):
        return self.score == other.score

    def __gt__(self, other):
        return self.score > other.score

    def __repr__(self):
        return f"Counts_to_Key(key:{self.key}, score:{self.score})"


@CACHE_STRATEGY_REGISTRY.register()
class LFU_Strategy(Strategy_Base):
    """
        删除访问频率最低的部分
        drop items with smaller counts
    """

    name = ":by_counts:LFU_Strategy"

    def __init__(self):
        self.order_ls = list()  # 最小堆
        self.record_s = dict()

    def notified_by_write_of_cache(self, key, value, metadata):
        assert (metadata is None or metadata["counts"] == 0) and key not in self.record_s
        self.record_s[key] = Score_to_Key(score=0, key=key)
        heapq.heappush(self.order_ls, self.record_s[key])

    def notified_by_read_of_cache(self, key, value, metadata):
        assert metadata is None or metadata["counts"] == self.record_s[key].score + 1
        self.record_s[key].score += 1

    def notified_by_remove_of_cache(self, key, metadata):
        temp = self.record_s.pop(key)
        temp.score = -1
        heapq.heapify(self.order_ls)
        temp = heapq.heappop(self.order_ls)
        assert temp.key == key

    def notified_by_clear_of_cache(self):
        self.record_s.clear()
        self.order_ls.clear()

    def suggest(self, refactor_size):
        heapq.heapify(self.order_ls)
        res = []
        temp_ls = []
        for _ in range(len(self.order_ls) - refactor_size):
            temp = heapq.heappop(self.order_ls)
            res.append(temp.key)
            temp_ls.append(temp)
        self.order_ls = temp_ls + self.order_ls
        return res

    def clear(self):
        self.notified_by_clear_of_cache()


# 添加其他别名
for name in [":by_counts:LFU", ":by_counts:Least_Frequently_Used", ":by_counts:drop_smaller"]:
    CACHE_STRATEGY_REGISTRY.add(obj=LFU_Strategy, name=name, b_force=False, b_execute_now=False)

if __name__ == "__main__":
    max_heap = []
    for i, j in [(-5, "a"), (-2, "adafd"), (-7, "ad"), (-1, "a"), (-10, "fasfde")]:
        heapq.heappush(max_heap, Score_to_Key(i, j))
    print(max_heap)



from .variable import Table_Format
from .get_format import get_format
from .padding_misaligned_values import padding_misaligned_values
from .convert_format import convert_format
from .find_tables import find_tables
from .generate_table import generate_table



import numpy as np
from scipy.stats import chi2
from kevin_toolbox.patches.for_numpy.linalg import normalize
from kevin_toolbox.patches.for_numpy.random import get_rng
from kevin_toolbox.patches.for_numpy.random.variable import DEFAULT_SETTINGS
from kevin_toolbox.computer_science.algorithm.cache_manager import Cache_Manager

cache_for_cov = Cache_Manager(upper_bound=20, refactor_size=0.5)


def truncated_multivariate_normal(
        mean, cov=None, low_radius=None, high_radius=None, size=None, b_check_cov=False,
        hit_ratio_threshold=DEFAULT_SETTINGS["truncated_multivariate_normal"]["hit_ratio_threshold"],
        expand_ratio=DEFAULT_SETTINGS["truncated_multivariate_normal"]["expand_ratio"],
        **kwargs):
    """
        从截断的多维高斯分布中进行随机采样

        参数：
            mean:                   <list of float> 均值
            cov:                    <matrix> 协方差矩阵
            low_radius,high_radius: <float> 截断边界
                                        注意：
                                            - 是截断多少倍 sigma 以内的部分。
                                            - low_radius 是指排除该半径距离内的点
                                            - high_radius 是指包含该半径距离内的点
                                            - 区间为左闭右开，亦即 [low_radius, high_radius)
            size:                   <tuple/list/int/None> 输出的形状
            b_check_cov:            <boolean> 是否检查 cov 是正半定矩阵。
                                        默认为 False 此时不检查
                                        当设置为 True 时，若不通过检查将报错。

        用于调节采样效率的超参数（与设备情况有关）：
            hit_ratio_threshold:    <float> 决定采样方式的阈值
                                        当 hit_ratio 小于该阈值时，使用方式 2 （重要性采样）来生成，
                                        当大于阈值时，使用方式 1 采样 expand_ratio * size 个样本再挑选符合落在截断区间内的样本
            expand_ratio:           <float> 方式1的系数
                                        要求大于 1

        其他参数：
            seed:                   <int> 随机种子
            rng:                    <Random Generator> 给定的随机采样器
                            以上参数二选一

        返回:
            res:                    当 shape 为 None 时，返回的是与mean大小相同的 n 维向量
                                    否则返回的是 shape+[len(mean)] 维度的张量
    """
    # 检查参数
    assert len(mean) > 1
    assert high_radius is None or 0 < high_radius
    assert low_radius is None or 0 <= low_radius
    if high_radius is not None and low_radius is not None:
        assert low_radius < high_radius
    if b_check_cov and cov is not None:
        cov = np.asarray(cov)
        assert np.allclose(cov, cov.T) and np.all(np.linalg.eig(cov)[0] > 0)
    #
    rng = get_rng(**kwargs)
    low = None if low_radius is None else low_radius ** 2
    high = None if high_radius is None else high_radius ** 2

    # quick return
    if (low is None or low == 0) and high is None:
        return rng.multivariate_normal(mean, cov, size=size, check_valid="warn" if b_check_valid else "ignore")

    # 因为标准高维高斯分布的采样点的方向服从均匀分布，而距离服从自由度为k的卡方分布
    #   因此可以把方向和距离分开来进行采样
    raw_size = 1 if size is None else np.prod([size])

    # 对方向进行采样
    theta = rng.normal(0, 1, size=[raw_size, len(mean)])
    theta = normalize(v=theta, ord=2, axis=-1)

    # 对距离进行采样
    # 计算命中概率
    cdf_high = chi2.cdf(high, df=len(mean)) if high is not None else 1
    cdf_low = chi2.cdf(low, df=len(mean)) if low is not None else 0
    hit_prob = cdf_high - cdf_low
    if hit_prob >= hit_ratio_threshold:
        # 采样方式1
        delta = np.empty(raw_size)
        count = 0
        while count < raw_size:
            temp = rng.chisquare(len(mean), int((raw_size - count) / hit_prob * expand_ratio) + 1)
            if low is not None:
                temp = temp[temp >= low]
            if high is not None:
                temp = temp[temp < high]
            delta[count:count + len(temp)] = temp[:raw_size - count]
            count += len(temp)
    else:
        # 采样方式2（重要性采样）
        # 从均匀分布中采样
        delta = rng.uniform(cdf_low, cdf_high, raw_size)
        # 对均匀分布的样本进行逆变换得到截断正态分布的样本
        delta = chi2.ppf(delta, df=len(mean))

    # 整合方向和距离
    res = theta * delta[:, None] ** 0.5

    # 根据协方差矩阵进行缩放
    if cov is not None:
        A = cache_for_cov.get(cov.tobytes(), default_factory=lambda: np.linalg.cholesky(cov), b_add_if_not_found=True)
        res = (A @ res.T).T
    res += mean

    if size is None:
        res = res[0]
    else:
        size = [size] if isinstance(size, int) else list(size)
        res = res.reshape(size + [len(mean)])

    return res


if __name__ == '__main__':
    print(truncated_multivariate_normal(mean=[0, 0], cov=np.array([[1, 0.5], [0.5, 1]]),
                                        high_radius=2, seed=114, size=[2, 5]))

    # 可视化
    import matplotlib.pyplot as plt

    res_ = truncated_multivariate_normal(mean=[0, 0], cov=np.array([[1, 0.5], [0.5, 1]]), low_radius=0.5,
                                         high_radius=1.5, seed=114, b_check_cov=True,
                                         size=10000)
    plt.scatter(res_[:, 0], res_[:, 1], c='r', s=1)

    plt.show()



import os
from kevin_toolbox.computer_science.algorithm.registration import Registry

ignore_s = [
    {
        "func": lambda _, __, path: os.path.basename(path) in ["temp", "test", "__pycache__",
                                                               "_old_version"],
        "scope": ["root", "dirs"]
    },
]

# 包含缓存管理or更新策略
CACHE_STRATEGY_REGISTRY = Registry(uid="CACHE_STRATEGY_REGISTRY")

CACHE_STRATEGY_REGISTRY.collect_from_paths(
    path_ls=[os.path.join(os.path.dirname(__file__), "strategy"), ],
    ignore_s=ignore_s,
    b_execute_now=False
)

# 包含缓存构建器
CACHE_BUILDER_REGISTRY = Registry(uid="CACHE_BUILDER_REGISTRY")

CACHE_BUILDER_REGISTRY.collect_from_paths(
    path_ls=[os.path.join(os.path.dirname(__file__), "cache"), ],
    ignore_s=ignore_s,
    b_execute_now=False
)



from kevin_toolbox.patches.for_streamlit.markdown import show_table


def show(text, doc_dir=None):
    """
        st.markdown 的改进版，具有以下优点
            - 对于带有图片的表格，用分列分行显示
            - 能够正确显示本地的图片
    """
    show_table(text=text, doc_dir=doc_dir)



from collections import OrderedDict
from kevin_toolbox.computer_science.algorithm.cache_manager.strategy import LRU_Strategy
from kevin_toolbox.computer_science.algorithm.cache_manager.variable import CACHE_STRATEGY_REGISTRY


@CACHE_STRATEGY_REGISTRY.register()
class FIFO_Strategy(LRU_Strategy):
    """
        删除最后一次访问时间最久远的部分
        drop items with smaller initial_time
    """

    name = ":by_initial_time:FIFO_Strategy"

    def notified_by_read_of_cache(self, key, value, metadata):
        pass


# 添加其他别名
for name in [":by_initial_time:FIFO", ":by_initial_time:First_In_First_Out", ":by_initial_time:drop_smaller"]:
    CACHE_STRATEGY_REGISTRY.add(obj=FIFO_Strategy, name=name, b_force=False, b_execute_now=False)
